import itertools
import random

import numpy as np

from multivar_horner.global_settings import FLOAT_DTYPE, UINT_DTYPE


def proto_test_case(data, fct):
    all_good = True
    for input, expected_output in data:
        print('\n')
        actual_output = fct(input)
        print(f'p({input[2]}) == {expected_output}')
        if actual_output != expected_output:
            print(f'ERROR: p(x) == {actual_output}')
            all_good = False
        else:
            print('OK.')

    assert all_good


def get_rnd_poly_properties(all_exponents, degree, max_abs_coeff=1.0, integer_coeffs=False, enforce_degree=False):
    """
    for meaningful test results uniformly distributed random test polynomials are required
    -> every possible polynomial should appear with equal probability
    <=> every 'sparsity' AND every possible monomial have equal probability

    sparsity: the ratio of 'active' monomials out of all possible monomials

    there must be at least 1 entry ('active' monomial)

    :param all_exponents:
    :param degree: the total degree the monomial must (at least) have
    :param max_abs_coeff:
    :param integer_coeffs:
    :return: evenly distributed settings of a polynomial with the given possible exponents
    """

    max_degree_monomial_idxs = all_exponents.sum(axis=1) >= degree
    assert np.any(max_degree_monomial_idxs)

    # every exponent can take the values in the range [0; maximal_degree]
    max_nr_exponent_vects = all_exponents.shape[0]

    # pick a random sparsity: controlled by the uniformly distributed probability a monomial is present
    activation_thres = random.random()
    row_activation = np.random.rand(max_nr_exponent_vects)
    row_idxs = row_activation > activation_thres

    if enforce_degree:
        # the settings must correspond to a polynomial of at least the requested total degree
        # -> at least one monomial of this total degree (or higher) must be active
        while not np.any(row_idxs & max_degree_monomial_idxs):
            row_activation = np.random.rand(max_nr_exponent_vects)
            row_idxs = row_activation > activation_thres
    else:
        if np.sum(row_idxs) == 0:  # there must be at least 1 entry ('active' monomial)
            row_idxs[random.randint(0, max_nr_exponent_vects - 1)] = 1

    exponents = all_exponents[row_idxs, :]
    nr_monomials = exponents.shape[0]
    coefficients = (np.random.rand(nr_monomials, 1) - 0.5) * (2 * max_abs_coeff)  # [ -max_abs_coeff; max_abs_coeff]
    coefficients = coefficients.astype(FLOAT_DTYPE)
    if integer_coeffs:
        coefficients = np.round(coefficients)

    return coefficients, exponents


def all_possible_exponents(dim, deg):
    """
    generate a fully occupied exponent matrix for a polynomial of lp_degree = infinity
    -> the given degree is the total degree of the polynomials
    :param dim:
    :param deg:
    :return:
    """

    # def cntr2exp_vect(cntr):
    #     exp_vect = np.empty(dim, dtype=UINT_DTYPE)
    #     for d in range(dim - 1, -1, -1):
    #         divisor = (deg + 1) ** d
    #         # cntr = quotient*divisor + remainder
    #         quotient, remainder = divmod(cntr, divisor)
    #         exp_vect[d] = quotient
    #         cntr = remainder
    #     return exp_vect

    max_nr_exponent_vects = (deg + 1) ** dim
    all_possible = np.empty((max_nr_exponent_vects, dim), dtype=UINT_DTYPE)

    # for i in range(max_nr_exponent_vects):
    #     all_possible[i] = cntr2exp_vect(i)

    for i, exponents in enumerate(itertools.product(range(deg + 1), repeat=dim)):
        all_possible[i] = np.asarray(exponents)

    # there must not be duplicate exponent vectors
    assert all_possible.shape == np.unique(all_possible, axis=0).shape

    return all_possible


def rnd_settings_list(length, dim, degree, *args, **kwargs):
    all_exponent_vect = all_possible_exponents(dim, degree)
    settings_list = [get_rnd_poly_properties(all_exponent_vect, degree, *args, **kwargs) for i in range(length)]

    # # random settings should have approx. half the amount of maximal entries on average
    # num_monomial_entries = 0
    # for settings in settings_list:
    #     num_monomial_entries += settings[0].shape[0]
    #
    # avg_monomial_entries = num_monomial_entries / length
    # max_monomial_entries = int((maximal_degree + 1) ** dim)
    # print(avg_monomial_entries, max_monomial_entries)
    return settings_list


def rnd_input_list(length, dim, max_abs_val):
    return [(np.random.rand(dim) - 0.5) * (2 * max_abs_val) for i in range(length)]


def vectorize(obj):
    return lambda x: np.array([obj(np.atleast_1d(x_i)) for x_i in x])


# vectorised version of naive_eval() in multivar_horner.helpers_fcts_numba
def naive_eval_reference(X, exponents, coefficients):
    return np.dot(np.array([[(x ** ex).prod() for ex in exponents] for x in X]), coefficients)



import numpy as np

from .global_settings import ID_ADD, ID_MULT, UINT_DTYPE
from .helper_classes import PriorityQueue2D
from .helpers_fcts_numba import compile_valid_options, count_num_ops, count_usage, factor_num_ops, num_ops_1D_horner


class FactorisationNode:
    """
    A node representing a factorisation of a polynomial:
    p = f_1 * p_1 + p_2
    its sub problems are polynomials as well ('divide and conquer' -> recursion)
    factorisation in this way results in building a binary "Horner Factorisation Tree"
    TODO factorisation is ambiguous: f = f1 f2 = f2 f1
        increases the amount of possible factorisations (relevant for search over all factorisations!)
        allow whole monomials as factors or remove ambiguity by uniquely identifying identical polynomials
        just important for optimal factorisation search. not a priority atm.
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['factor', 'node1_fact', 'node2', 'factorized_rows', 'non_factorized_rows', 'value_idx']

    def __init__(self, factor, node1_fact, node2, factorized_rows, non_factorized_rows):
        self.factor = factor
        self.node1_fact = node1_fact
        self.node2 = node2  # can be None if all monomials shared the factor
        self.factorized_rows = factorized_rows
        self.non_factorized_rows = non_factorized_rows
        self.value_idx = None

    def get_string_representation(self, *args, **kwargs):
        s = self.factor.__str__(*args, **kwargs)
        s += ' ({})'.format(
            self.node1_fact.get_string_representation(*args, **kwargs))
        if self.node2 is not None:
            s += ' + {}'.format(
                self.node2.get_string_representation(*args, **kwargs))
        return s

    def __str__(self, *args, **kwargs):
        return self.get_string_representation(*args, **kwargs)

    def __gt__(self, other):  # arbitrary, required for sorting in heap
        return True

    def compile_factors(self, factor_container, coefficient_idxs):
        # create and store the unique ScalarFactor instance
        # own factor is a scalar factor: only one property
        self.factor = factor_container.get_factor([self.factor])

        coeff_idxs1 = coefficient_idxs[self.factorized_rows]
        self.node1_fact.compile_factors(factor_container, coeff_idxs1)

        if self.node2 is not None:
            coeff_idxs2 = coefficient_idxs[self.non_factorized_rows]
            self.node2.compile_factors(factor_container, coeff_idxs2)

        # this node does not need its own index in the value array, but can reuse the index of its first node
        # child nodes are being evaluated earlier
        # NOTE: all nodes have at least one value index, since a node always contains a coefficient
        self.value_idx = self.node1_fact.value_idxs[0]

    def get_recipe(self):
        # p = f_1 * p_1 + p_2
        # values[target] = values[target] *op* values[source]

        # IMPORTANT: sub trees have to be evaluated BEFORE their values can be used!
        # -> add their recipes first
        tree_recipe, op_recipe = self.node1_fact.get_recipe()

        # the value at values[node.idx] is the evaluated value of this node
        tree_recipe += [
            # instruction encoding: target, source
            # multiply the value of the node1 with the value of the factor
            (self.node1_fact.value_idxs[0], self.factor.value_idx),
        ]
        # separate: op (binary: 0/1)
        op_recipe += [ID_MULT]

        if self.node2 is not None:
            tree_recipe_sub, op_recipe_sub = self.node2.get_recipe()
            tree_recipe += tree_recipe_sub
            op_recipe += op_recipe_sub

            tree_recipe += [
                # add the value of node2 to this value
                (self.node1_fact.value_idxs[0], self.node2.value_idxs[0]),
            ]
            op_recipe += [ID_ADD]

        return tree_recipe, op_recipe


class OptimalFactorisationNode(FactorisationNode):
    __slots__ = ['cost_estimate', 'factorisation_measure', 'fully_factorized']

    def __init__(self, factor, node1_fact, node2, factorized_rows, non_factorized_rows):
        super(OptimalFactorisationNode, self).__init__(factor, node1_fact, node2, factorized_rows, non_factorized_rows)
        self.cost_estimate: int = 0
        # IDEA: when different factorisations have the same cost estimate,
        # favour the one which is factorised the most already
        self.factorisation_measure = 0
        self.fully_factorized = False
        self.update_properties()

    def update_properties(self):
        self.cost_estimate = factor_num_ops(*self.factor) + self.node1_fact.cost_estimate
        self.factorisation_measure = self.node1_fact.factorisation_measure

        if self.node2 is not None:
            self.cost_estimate += self.node2.cost_estimate
            self.factorisation_measure += self.node2.factorisation_measure
            self.fully_factorized = self.node1_fact.fully_factorized and self.node2.fully_factorized
        else:
            self.fully_factorized = self.node1_fact.fully_factorized

    def refine(self):
        assert not self.fully_factorized
        self.node1_fact.refine()
        if self.node2 is not None:
            self.node2.refine()
        self.update_properties()


class BasePolynomialNode:
    """
    The base class representing a multivariate polynomial as a child node in a factorisation tree for finding
        a good factorisation of a bigger multivariate polynomial

    TODO document
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['exponents', 'unique_exponents', 'num_monomials', 'dim', 'children', 'value_idxs',
                 'factors', 'has_children', 'children_class', 'factorisation_class']

    def __init__(self, exponents, *args, **kwargs):
        self.exponents = exponents
        self.num_monomials = self.exponents.shape[0]
        self.dim = self.exponents.shape[1]

        self.unique_exponents = []

        #  all monomials are independent...
        # if np.all(np.sum(exponent_matrix, axis=0) == np.max(exponent_matrix, axis=0)):
        # NOTE: not useful (too expensive)

        for d in range(self.dim):
            dim_exponents = self.exponents[:, d]
            # sort in ascending order (used for finding the most common scalar factor)
            dim_exponents_unique = np.unique(dim_exponents)
            dim_exponents_unique.sort()
            # exponent 0 is never a valid factorisation option
            if dim_exponents_unique[0] == 0:
                self.unique_exponents.append(dim_exponents_unique[1:])
            else:
                self.unique_exponents.append(dim_exponents_unique)

        self.children = None
        self.value_idxs = None
        self.factors = None
        self.has_children = False
        self.children_class = None
        self.factorisation_class = None
        self.post_init()

    def post_init(self):
        self.children_class = BasePolynomialNode
        self.factorisation_class = FactorisationNode

        # fully factorize automatically
        option = self.choose_option()
        if option is None:
            self.has_children = False
        else:
            self.has_children = True
            self.factorize(*option)

    def choose_option(self):
        """
        :return: the option with the maximum usage in all monomials (<- 'heuristic')
        """

        if self.num_monomials <= 1:
            # there are no options when only one monomial is left in a node!
            return None

        max_usage = 0
        max_usage_option = None

        for dim, dim_unique_exponents in enumerate(self.unique_exponents):
            # of all existing exponents, the smallest ones for sure have the highest usage
            for exp in dim_unique_exponents:
                usage = count_usage(dim, exp, self.exponents)
                if usage > max_usage:
                    max_usage = usage
                    max_usage_option = (dim, exp)
                # no other exponent in this dimension can have a higher usage
                break

        return max_usage_option

    def get_string_representation(self, coefficients=None, coeff_fmt_str='{:.2}', factor_fmt_str='x_{dim}^{exp}'):

        if self.has_children:
            return self.get_child().get_string_representation(coefficients=coefficients, coeff_fmt_str=coeff_fmt_str,
                                                              factor_fmt_str=factor_fmt_str)
        else:
            monomial_representations = []
            for i, exp_vect in enumerate(self.exponents):
                if coefficients is None:
                    coeff_repr = 'c'
                else:
                    coeff_idx = self.value_idxs[i]  # look up the correct index of the coefficient
                    coeff_repr = coeff_fmt_str.format(coefficients[coeff_idx, 0])

                monomial_repr = [coeff_repr]
                for dim, exp in enumerate(exp_vect):
                    if exp > 0:
                        monomial_repr.append(factor_fmt_str.format(**{'dim': dim + 1, 'exp': exp}))

                monomial_representations.append(' '.join(monomial_repr))
            return ' + '.join(monomial_representations)

    def __str__(self):
        return self.get_string_representation()

    def factorize(self, dim, exp):
        """
        factorize the polynomial represented by this node:
        find all the monomials 'benefiting' from this factorisation
        create and store a factorized representation of this polynomial
        'top down' approach
        """

        factorized_rows = self.exponents[:, dim] >= exp
        exponents1_fact = self.exponents[factorized_rows, :]
        # the factor has to be deducted from the exponents (="factored out")
        exponents1_fact[:, dim] -= exp

        node1_fact = self.children_class(exponents1_fact)

        # create a representation of the polynomial with all the remaining monomials
        non_factorized_rows = np.invert(factorized_rows)
        if not np.any(non_factorized_rows):
            # no monomials remain, do not create a polynomial representation
            node2 = None
        else:
            exponents2 = self.exponents[non_factorized_rows, :]
            node2 = self.children_class(exponents2)

            # if DEBUG:  # TODO move to tests, create unit tests
            #     assert exp > 0
            #     assert not np.any(exponents1_fact < 0)
            #     assert not np.any(exponents2 < 0)
            #     assert exponents1_fact.shape[1] == exponents2.shape[1] and len(exponents2.shape) == 2
            #     assert exponents1_fact.shape[0] + exponents2.shape[0] == self.num_monomials

        factor = (dim, exp)
        child = self.factorisation_class(factor, node1_fact, node2, factorized_rows, non_factorized_rows)
        self.store_child(child)

    def store_child(self, child):
        self.children = child  # allow only one factorisation

    def get_child(self):
        return self.children

    def compile_factors(self, factor_container, coefficient_idxs):
        """
        factorisation has been done (fixed)
        now "collect" all existing factors and link value addresses correctly
        'bottom up' approach
        :param factor_container: class storing and managing all existing factors
        :param coefficient_idxs: the indices of the given coefficients in the global coefficient array
            of the root polynomial
        :return:
        """
        if self.has_children:
            child = self.get_child()
            child.compile_factors(factor_container, coefficient_idxs)
            # this node does not need its own index (storage space) in the value array,
            # it can reuse the value index of its sub problem
            self.value_idxs = [child.value_idx]
        else:
            # this node cannot be factorized (no options)
            # for evaluation the sum of all evaluated monomials multiplied with their coefficients has to be computed
            # p = c1 * mon1 + c2 * mon2 ...
            # create factors representing the remaining monomials
            self.factors = factor_container.get_factors(self.exponents)  # retains the ordering!
            # remember where in the coefficient array the coefficients of this polynomial are being stored
            self.value_idxs = coefficient_idxs

    def get_recipe(self):
        """
        :return: the list of instructions ("recipe") required for evaluating the represented factorisation
        """
        if self.has_children:
            # the own recipe is the recipe of its factorisation
            return self.get_child().get_recipe()
        else:
            # this node has not been factorized (represents a regular polynomial)
            # for evaluation, the sum of all evaluated monomials (='factors')
            # multiplied with their coefficients has to be computed
            # p = c1 * mon1 + c2 * mon2 ...
            # the value of the coefficients in the value array are only being used once
            # -> their address can be reused for storing intermediary results
            # the final evaluated value of this node must be stored at the first of its value indices
            # this
            initial_coeff_value_idx = self.value_idxs[0]
            factor = self.factors[0]
            tree_recipe = []
            op_recipe = []

            if factor is not None:
                # multiply the first coefficients with the value of the first factor
                tree_recipe += [(initial_coeff_value_idx, self.factors[0].value_idx)]
                op_recipe += [ID_MULT]

            for i in range(1, len(self.value_idxs)):
                coeff_value_idx = self.value_idxs[i]
                factor = self.factors[i]
                if factor is not None:
                    # multiply each coefficients with the value of the factor
                    tree_recipe += [
                        (coeff_value_idx, factor.value_idx),
                    ]
                    op_recipe += [ID_MULT]

                # add this value to the previously computed value
                tree_recipe += [
                    (initial_coeff_value_idx, coeff_value_idx),
                ]
                op_recipe += [ID_ADD]
            return tree_recipe, op_recipe


class OptimalPolynomialNode(BasePolynomialNode):
    """
    A class representing a multivariate polynomial with all possible factorisation options.
    The different factorisation options are being stored sorted in a heap (self.children)
    An instance of this class is a node (=subproblem) in search tree for finding
    the optimal factorisation of a bigger multivariate polynomial
    """

    # prevent dynamic attribute assignment (-> safe memory)
    __slots__ = ['options', 'cost_estimate', 'factorisation_measure', 'fully_factorized']

    def __init__(self, exponents, *args, **kwargs):

        self.options = None
        self.cost_estimate = 0
        self.factorisation_measure = 0
        self.fully_factorized = True

        # self.post_init() is being called at the end
        super(OptimalPolynomialNode, self).__init__(exponents, *args, **kwargs)

    def post_init(self):
        self.children_class = OptimalPolynomialNode
        self.factorisation_class = OptimalFactorisationNode
        self.get_all_options()
        self.estimate_cost()

    def __eq__(self, other):
        """
        TODO improvement:
        different orders of factorisations can lead to identical sub problems
        solving them independently leads to unnecessary computations
        -> before creating a new node check if it already exists (equality comparison required)
        equality test can be aborted when certain global properties of nodes do not match
        (cost_estimate, dimensions of exponent matrix...)
        only actually compare exponent matrices for equality after all previous tests passed
        ATTENTION: after changing a multiply used node all parents have to be updated!
        """
        raise NotImplementedError

    def store_child(self, child):
        self.children.put(child)

    def get_child(self):
        return self.children.get()  # return the currently most promising factorisation

    def estimate_cost(self):
        """
        :return: the estimated amount of operations required needed when evaluating this polynomial in fully
            factorized form. true amount when fully factorized otherwise a lower bound.

        Every existing factor (represented by exponents) has to be evaluated, which minimally
        requires the amount of operations all respective 1D-Horner factorisations would take.
        NOTE: not counting the amount of ADD operations and operations needed because of the coefficients,
            since they stay constant in sum irrespective of the factorisation (= number of monomials)

        TODO: improvement idea: more ops certainly needed when exponents appear multiple times in exponent vectors
         (monomials are unique) and can hence by definition not be fully factorized... but do not count multiple times
         -> derive rule.
        """
        if self.fully_factorized:
            # the actual operation count can be computed:
            self.cost_estimate = count_num_ops(self.exponents)
        else:
            # count one multiplication with the coefficients for each monomial
            heuristic = np.count_nonzero(np.any(self.exponents, axis=1))
            for dim_unique_exp in self.unique_exponents:
                heuristic += num_ops_1D_horner(dim_unique_exp)
            self.cost_estimate = heuristic

    def get_all_options(self):
        """
        :return: a list of all meaningful options (dim, exp) for factorisation

        NOTE: scalar factors (<-> exponents) which appear only in one monomial are not meaningful factors
            Picking them would only lead to separating this one monomial = equal to not factorising at all
            no decrease in #ops -> the heuristic would not change -> also no change in search
            monomials without any common exponents should not be factorized (can be computed without factorisation)
        """

        options = []
        if self.num_monomials <= 1:
            # there are no options when only one monomial is left in a node!
            return options

        for dim, dim_unique_exponents in enumerate(self.unique_exponents):
            usage_vector = np.zeros(dim_unique_exponents.shape, dtype=UINT_DTYPE)
            valid_option_vector = np.zeros(dim_unique_exponents.shape, dtype=bool)
            # TODO test
            valid_option_vector = compile_valid_options(dim, valid_option_vector, usage_vector, dim_unique_exponents,
                                                        self.exponents)

            for exp in dim_unique_exponents[valid_option_vector]:
                options.append((dim, exp))

        self.factorisation_measure = len(options)
        self.fully_factorized = len(options) == 0
        self.options = options

    def refine(self):
        """
        factorize further (grow tree)

        TODO improvement:
        initially compute good factorisation (with heuristic). use as upper bound for the #ops
        do not keep factorisations which have a higher or !equal! estimated #ops (<- save memory)
        when heap becomes empty, the heuristic solution is optimal
        apply to all layers, pass upper bound down the tree (reduce accordingly)
        NOTE: when searching all optimal solutions having the same #ops as the upper bound is allowed!
        ATTENTION: cost estimate is not counting additions (cf. estimate_cost() ) do not compare with total #ops!

        TODO improvement:
        while loop keep factorising until no other option has a lower heuristic or fully factorised
        -> check what the next lowest heuristic is. corner case: no other option available!
        """
        # TODO remove
        # print('\nrefining', self.fully_factorized, self.cost_estimate, self)
        if self.fully_factorized:
            # this check is required, because nodes at the same depth in the factorisation tree might have different
            # factorisation statuses (.refine() is still being called)
            return

        if self.has_children:
            # the heap has been initialized
            # pick the most promising factorisation subtree, factorise it further
            promising_node = self.children.pop()  # pop, not get!
            # print('refining', promising_node.fully_factorized,  promising_node.cost_estimate, promising_node)
            promising_node.refine()
            # print('done', promising_node.fully_factorized,  promising_node.cost_estimate, promising_node)
            self.store_child(promising_node)
        else:
            # no factorisation has been done before
            # initialize heap
            # the cost estimate of less factorised bad factorisations and more factorised good factorisation is similar
            # to reach a fully factorised solution faster, more factorised options have to be favoured
            # -> have a "2D heap" sorted first after the cost estimate and then after the factorisation progress
            self.children = PriorityQueue2D()
            # use all options and store the different factorisation results in heap
            # there might be many options and trying all is often unnecessary,
            # but trying just one is also not efficient, because the unfactorized node most certainly
            # has a lower cost estimate and hence would get picked next anyway
            for dim, exp in self.options:
                self.factorize(dim, exp)

            self.has_children = True

        # print('all children:')
        # for c in self.children.get_all():
        #     print('\t', c.fully_factorized, c.cost_estimate,c)

        # the new properties of this node are equal to the properties of the most promising new subtree
        promising_node = self.get_child()
        # print('most promising:', promising_node.fully_factorized, promising_node.cost_estimate, promising_node)

        self.factorisation_measure = promising_node.factorisation_measure
        self.fully_factorized = promising_node.fully_factorized
        # cost_estimate_prev = self.cost_estimate
        self.cost_estimate = promising_node.cost_estimate
        # the initial heuristic is a lower bound for the true cost
        # the cost estimate should always increase or stay constant (= getting more realistic) with every refinement
        # assert cost_estimate_prev <= self.cost_estimate


class OptimalFactorisationRoot(OptimalPolynomialNode):
    """
    Class for finding optimal factorisations of given multivariate polynomials
    Basic idea: adapted A* search
        allow all meaningful possible factorisation while ranking them according to
        their lowest possible amount of operations needed for evaluation (heuristic)
        iteratively factorize the most promising factorisation further until a full factorisation has been found

    guaranteed to yield a factorisation with the minimal amount of operations required during evaluation (optimal)
    memory and time consumption is much higher than with using a heuristic to pick a single factorisation

    TODO improvement: analyse and optimize runtime of optimal factorisation search
    tradeoff between accuracy of the heuristic (-> ignore irrelevant factorisations)
        and time to compute heuristic (-> better to just try factorisations)


    TODO improvement idea: instead of just optimizing for the lowest amount of operations,
    also consider the different costs of the different operations on computer architectures
    (optimize total evaluation time).
    ATTENTION: a completely new heuristic for estimating the lower bound would be required!
    """

    def __init__(self, exponents, factor_container, *args, **kwargs):
        """
        :param coefficients: a numpy array column vector of doubles representing the coefficients of the monomials
        :param exponents: a numpy array matrix of unsigned integers representing the exponents of the monomials
            the ordering does not matter, but every exponent row has to be unique!
        :param rectify_input: whether to convert the input parameters into compatible numpy arrays
        :param validate_input: whether to check if the input parameters fulfill the requirements
        """
        super(OptimalFactorisationRoot, self).__init__(exponents, *args, **kwargs)

        while not self.fully_factorized:
            # TODO report current search status, verbosity option, not every loop
            # (#total factorisations done, #max depth, #open possibilities, upper bound, lower bound)
            self.refine()

        # the most promising factorisation (lowest cost estimate) is fully factorized now
        # collect all appearing factors in the best solution and assign ids (=idx in the value array) to the nodes
        # need to store which coefficient is being used where in the factorisation tree (coeff_id -> value_idx)
        coefficient_idxs = np.arange(self.num_monomials, dtype=np.int)
        self.compile_factors(factor_container, coefficient_idxs)

    def find_all_optimal(self):
        """
        TODO refine and collect all fully factorized optimal solutions until the next best factorisation
            has a higher cost estimate
        Attention: do not remove entire subtree, but just one realisation each time!
        TODO analyse obtained optimal factorisations
        use this data to derive better heuristics for factorizing more optimally a priori
        :return:
        """
        # optimal_factorisations = []
        raise NotImplementedError


class HeuristicFactorisationRoot(BasePolynomialNode):
    """
    Class for finding a good factorisation of a given multivariate polynomial by iteratively factoring out the most
    commonly used factor (heuristic).
    """

    def __init__(self, exponents, factor_container, *args, **kwargs):
        super(HeuristicFactorisationRoot, self).__init__(exponents, *args, **kwargs)
        # polynomial is fully factorized now
        # collect all appearing factors in the factorisation tree and assign ids (=idx in the value array) to the nodes
        # need to store which coefficient is being used where in the factorisation tree (coeff_id -> value_idx)
        coefficient_idxs = np.arange(self.num_monomials, dtype=np.int)
        self.compile_factors(factor_container, coefficient_idxs)

# TODO define factorisation taking the numerical stability (coefficients) into account



# -*- coding:utf-8 -*-

import numpy as np

from .global_settings import FLOAT_DTYPE, UINT_DTYPE


def rectify_coefficients(coefficients):
    rectified_coefficients = np.atleast_2d(np.asarray(coefficients, dtype=FLOAT_DTYPE)).reshape(-1, 1)
    return rectified_coefficients


def rectify_construction_parameters(coefficients, exponents):
    """
    convert the input into numpy arrays valid as input of AbstractPolynomial
    :param coefficients: possibly a python list of coefficients to be converted
    :param exponents: possibly a nested python list of exponents to be converted
     ATTENTION: when converting to unsigned integer, negative integers become large!
    :return: the input converted into appropriate numpy data types
    """
    rectified_coefficients = rectify_coefficients(coefficients)
    exponents = np.atleast_2d(exponents)
    if np.any(exponents < 0):
        raise ValueError('negative exponents are not allowed!'
                         f'the conversion to {UINT_DTYPE} would turn negative values into very large values')
    rectified_exponents = exponents.astype(UINT_DTYPE)
    return rectified_coefficients, rectified_exponents


def rectify_query_point(x):
    """
    convert the input into numpy ndarray valid as input to AbstractPolynomial.eval()
    :param x: possibly a python list to be converted
    :return: the input converted into appropriate numpy data type
    """
    rectified_x = np.atleast_1d(np.asarray(x, dtype=FLOAT_DTYPE))
    return rectified_x


def validate_coefficients(coefficients) -> None:
    if not isinstance(coefficients, np.ndarray):
        raise TypeError('coefficients must be given as numpy.ndarray')
    if coefficients.dtype.type is not FLOAT_DTYPE:
        raise TypeError(f'coefficients must have dtype {FLOAT_DTYPE}')
    if len(coefficients.shape) != 2 or coefficients.shape[1] != 1:
        raise ValueError('coefficients must be given as a [M, 1] ndarray')
    if coefficients.shape[0] == 0:
        raise ValueError('there must be at least one coefficient')

    # NOTE: "0 coefficients" are allowed because coefficients should be changeable!


def validate_construction_parameters(coefficients, exponents) -> None:
    validate_coefficients(coefficients)

    if not isinstance(exponents, np.ndarray):
        raise TypeError('exponents must be given as numpy.ndarray')
    if exponents.dtype.type is not UINT_DTYPE:
        # NOTE: this also ensures non negative exponents
        # value check is being done in rectification fct
        raise TypeError(f'exponents must have dtype {UINT_DTYPE}')
    if len(exponents.shape) != 2:
        raise ValueError('exponents must be given as a [M, N] ndarray (a list of exponent vectors)')
    if np.any(exponents < 0):
        raise ValueError('exponents must not be negative')
    if exponents.shape != np.unique(exponents, axis=0).shape:
        raise ValueError('there must not be duplicate exponent vectors')
    if coefficients.shape[0] != exponents.shape[0]:
        raise ValueError('there must be as many exponent vectors as coefficients')


def validate_query_point(x) -> None:
    if not isinstance(x, np.ndarray):
        raise TypeError('the query point x  must be given as numpy.ndarray')
    if x.dtype.type is not FLOAT_DTYPE:
        raise TypeError(f'the query point x must have dtype {FLOAT_DTYPE}')
    if len(x.shape) != 1:
        raise ValueError('the query point x must be given as a ndarray of shape [N]')



from typing import List, Union

import numpy

ID_MULT = False
ID_ADD = True  # ATTENTION: is being used in helpers_fcts_numba.py/eval_recipe()

# numba is expecting certain data types (static typing):
# INT_DTYPE = numpy.int64  # i8 =  8byte integer
UINT_DTYPE = numpy.uint32  # u4 =  4byte unsigned integer
FLOAT_DTYPE = numpy.float64  # f8 =  8byte float, default
BOOL_DTYPE = numpy.bool

# python typing
TYPE_1D_FLOAT = Union[numpy.ndarray, List[float]]
TYPE_2D_INT = Union[numpy.ndarray, List[List[int]]]

DEFAULT_PICKLE_FILE_NAME = 'multivar_polynomial.pickle'

DEBUG = False
# DEBUG = True



# -*- coding:utf-8 -*-
from .multivar_horner import HornerMultivarPolynomial, MultivarPolynomial

# https://docs.python.org/3/tutorial/modules.html#importing-from-a-package
# determines which objects will be imported with "import *"
__all__ = ('MultivarPolynomial', 'HornerMultivarPolynomial')



import heapq  # implementation of the heap queue algorithm, also known as the priority queue algorithm (binary tree)
from abc import ABC, abstractmethod
from typing import List


# modified sample code from https://www.redblobgames.com/pathfinding/a-star/
class PriorityQueue:
    def __init__(self):
        self.elements = []

    def __len__(self):
        return len(self.elements)

    def is_empty(self):
        return len(self.elements) == 0

    def put(self, item, cost):
        heapq.heappush(self.elements, (cost, item))

    def pop(self):
        # remove and returns (only) the item with the lowest cost
        return heapq.heappop(self.elements)

    def get(self):
        return self.elements[0]


class PriorityQueue2D:
    def __init__(self):
        self.lvl1_heap = PriorityQueue()
        self.cost2id = {}
        self.id2heap = {}

    def __len__(self):
        length = 0
        for lvl2_heap in self.id2heap.values():
            length += len(lvl2_heap)
        return length

    def is_empty(self):
        return self.lvl1_heap.is_empty()

    def put(self, item):
        cost1 = item.cost_estimate
        cost2 = item.factorisation_measure
        try:
            # look up the heap with this lvl1 cost
            heap_id = self.cost2id[cost1]
            lvl2_heap = self.id2heap[heap_id]
        except KeyError:
            # there exists no second level heap with this cost, create an empty one
            lvl2_heap = PriorityQueue()
            heap_id = id(lvl2_heap)
            self.cost2id[cost1] = heap_id
            self.id2heap[heap_id] = lvl2_heap
            # the lvl1 heap stores the ids of the corresponding heaps
            self.lvl1_heap.put(heap_id, cost1)

        # the lvl2 heaps store the actual items
        lvl2_heap.put(item, cost2)

    def pop(self):
        cost1, heap_id = self.lvl1_heap.get()
        lvl2_heap = self.id2heap[heap_id]
        cost2, item = lvl2_heap.pop()
        if lvl2_heap.is_empty():
            # remove entries referring to an empty heap
            self.lvl1_heap.pop()
            self.id2heap.pop(heap_id)
            self.cost2id.pop(cost1)
            del lvl2_heap

        return item

    def get(self):
        cost1, heap_id = self.lvl1_heap.get()
        lvl2_heap = self.id2heap[heap_id]
        cost2, item = lvl2_heap.get()
        return item  # return only the item without the costs

    def get_all(self):
        all_items = []
        for lvl2_heap in self.id2heap.values():
            for (cost, item) in lvl2_heap.elements:
                all_items.append(item)
        return all_items


class AbstractFactor(ABC):
    __slots__ = ['value_idx']

    @abstractmethod
    def __str__(self, *args, **kwargs):
        pass

    @abstractmethod
    def __repr__(self):
        pass

    @abstractmethod
    def compute(self, x, value_array):
        pass

    @abstractmethod
    def num_ops(self):
        pass

    @abstractmethod
    def get_recipe(self):
        pass

    def eval(self, value_array):
        """
        looks up the computed value in the value array
        self.compute() has to be called before!
        then the value is stored at the value index in the value array
        :param value_array:
        :return: the computed value of self
        """
        return value_array[self.value_idx]


class ScalarFactor(AbstractFactor):
    """
    a factor depending on just one variable: :math:`f(x) = x_d^e`
    """
    __slots__ = ['dimension', 'exponent']

    def __init__(self, factor_dimension, factor_exponent):
        self.dimension = factor_dimension
        self.exponent = factor_exponent
        self.value_idx = None  # initialize the idx with None to catch faulty evaluation tries

    def __str__(self, factor_fmt_str='x_{dim}^{exp}', *args, **kwargs):
        # NOTE: variable numbering starts with 1: x_1, x_2, ...
        # if self.exponent == 1:
        #     return 'x_{}'.format(self.dimension + 1)

        return factor_fmt_str.format(**{'dim': self.dimension + 1, 'exp': self.exponent})

    def __repr__(self, *args, **kwargs):
        return self.__str__(*args, **kwargs)

    def num_ops(self):
        return 1  # the number of instructions required for computing the value

    def compute(self, x, value_array):
        value_array[self.value_idx] = x[self.dimension] ** self.exponent

    def get_recipe(self):
        """
        for evaluation the input value has to be either copied or exponentiated
        :return: copy_recipe, scalar_recipe
        """
        if self.exponent == 1:
            # just copy x[dim] value
            # values[target] = x[source]
            return [(self.value_idx, self.dimension)], []
        else:
            # instruction encoding: target, source, exponent
            # values[target] = x[source] ** exponent
            return [], [(self.value_idx, self.dimension, self.exponent)]


class MonomialFactor(AbstractFactor):
    """
    a factor ('monomial') consisting of a product of scalar factors:
    :math:`m(x) = x_i^j * ... * x_k^l`

    Parameters
    ----------

    :param scalar_factors: a list of scalar factors the monomial consists of

    Attributes
    ----------

    factorisation_idxs: the indices of the values of all scalar 'sub' factors in the value array of the polynomial.
        cannot be set at construction time, because all scalar factors need to receive their index first,
        but not all required scalar factors might exist
    """
    __slots__ = ['scalar_factors', 'factorisation_idxs']

    def __init__(self, scalar_factors: List['ScalarFactor']):

        # assert len(scalar_factors) > 1, 'a monomial must consist of at least two scalar factors' # DEBUG
        self.scalar_factors = scalar_factors
        self.value_idx = None  # initialize the idx with None to catch faulty evaluation tries

    def __str__(self):
        return ' '.join([f.__str__() for f in self.scalar_factors])

    def __repr__(self):
        return self.__str__()

    def num_ops(self):
        # count the number of instructions done during compute (during eval() only looks up the computed value)
        return len(self.scalar_factors) - 1  # product of all scalar factors

    def compute(self, x, value_array):
        # IMPORTANT: compute() of all the sub factors has had to be called before!
        value = value_array[self.factorisation_idxs[0]]
        for idx in self.factorisation_idxs[1:]:
            value *= value_array[idx]

        value_array[self.value_idx] = value

    def get_recipe(self):
        """
        for evaluation all scalar factors have to be multiplied (= monomial_recipe)
        :return: monomial_recipe
        """
        # target = source1 * source2
        # instruction encoding: target, source1, source2
        target, source1, source2 = self.value_idx, self.factorisation_idxs[0], self.factorisation_idxs[1]
        monomial_recipe = [(target, source1, source2)]

        source1 = target  # always take the previously computed value
        for source2 in self.factorisation_idxs[2:]:
            # and multiply it with the remaining factor values
            monomial_recipe += [(target, source1, source2)]

        return monomial_recipe


class FactorContainer:
    """
    a class for storing and reusing all factors appearing in a factorisation
    """

    def __init__(self):
        self.scalar_factors = []
        self.monomial_factors = []
        self.property2idx_scalar = {}
        self.property2idx_monomial = {}

    def get_factor(self, property_list):
        """
        creates and stores objects of all required (sub-)factors if necessary
        :param property_list: a list of dimension and exponent tuples [(d1,e1), (d2,e2)...]
            representing the scalar factors of a monomial
        :return: the object representing the factor with the given properties
        """

        assert len(property_list) > 0  # TODO DEBUG
        # create all required scalar factors
        scalar_factors = []
        monomial_id = 1
        for prop in property_list:
            d, e = prop
            try:
                scalar_idx = self.property2idx_scalar[prop]
                scalar_factor = self.scalar_factors[scalar_idx]
            except KeyError:
                scalar_factor = ScalarFactor(d, e)
                self.property2idx_scalar[prop] = len(self.scalar_factors)
                self.scalar_factors.append(scalar_factor)

            scalar_factors.append(scalar_factor)

        assert len(scalar_factors) == len(property_list)  # TODO DEBUG

        if len(scalar_factors) == 1:
            # the requested factor is a scalar factor
            return scalar_factors[0]

        # the requested factor is a monomial consisting of multiple factors
        monomial_id = tuple(property_list)
        try:
            monomial_idx = self.property2idx_monomial[monomial_id]
            monomial_factor = self.monomial_factors[monomial_idx]
        except KeyError:
            monomial_factor = MonomialFactor(scalar_factors)
            self.property2idx_monomial[monomial_id] = len(self.monomial_factors)
            self.monomial_factors.append(monomial_factor)

        return monomial_factor

    def get_factors(self, exponent_matrix):
        """
        :param exponent_matrix:
        :return: a list of all factors represented by the exponent_matrix with an identical ordering
        """

        all_factors = []
        # TODO optimise, modularise into functions: monomial factor, scalar factor... use numpy routines!?
        for mon_nr in range(exponent_matrix.shape[0]):
            exponent_vector = exponent_matrix[mon_nr, :]
            scalar_factors = []
            property_list = []
            for dim, exp in enumerate(exponent_vector):
                if exp > 0:
                    prop = (dim, exp)
                    property_list.append(prop)
                    scalar_factor = self.get_factor([prop])
                    scalar_factors.append(scalar_factor)

            if len(scalar_factors) == 0:
                factor = None  # monomial consists of no factors
            elif len(scalar_factors) == 1:
                factor = scalar_factors[0]  # monomial consists of a single scalar factor
            else:
                factor = self.get_factor(property_list)  # monomial consists of a multiple scalar factors

            all_factors.append(factor)

        return all_factors



import pickle
import sys
import unittest
from itertools import product

import numpy as np

from multivar_horner import HornerMultivarPolynomial, MultivarPolynomial
from multivar_horner.global_settings import FLOAT_DTYPE
from tests.test_helpers import rnd_settings_list
from tests.test_settings import (
    DEGREE_RANGE, DIM_RANGE, DTYPE_HIGH_PREC, MAX_COEFF_MAGNITUDE,
    MAX_NUMERICAL_ERROR, NR_COEFF_CHANGES, NR_TEST_POLYNOMIALS, TEST_RESULTS_PICKLE,
)


def evaluate_numerical_error(dim, max_degree):
    # basic idea: evaluating a polynomial at x = all 1 should give the sum of coefficients
    # -> any deviation is the numerical error
    results = []
    x = np.ones(dim, dtype=FLOAT_DTYPE)
    max_error = 0.0
    ctr_total = 0
    ctr_total_max = NR_TEST_POLYNOMIALS * NR_COEFF_CHANGES

    print(f'evaluating numerical error: dim: {dim}, max. degree: {max_degree} ...')
    for poly_ctr, (coefficients, exponents) in enumerate(rnd_settings_list(NR_TEST_POLYNOMIALS, dim, max_degree,
                                                                           max_abs_coeff=MAX_COEFF_MAGNITUDE,
                                                                           integer_coeffs=False)):
        # debug: validate_input=True
        nr_monomials = exponents.shape[0]
        # find factorisation (expensive)
        poly_horner = HornerMultivarPolynomial(coefficients, exponents, validate_input=True)
        poly = MultivarPolynomial(coefficients, exponents)

        # # for evaluating only the polynomial properties without the numerical errors:
        # result = (poly, poly_horner, 0.0,0.0,0.0)
        # results.append(result)
        # continue

        coefficients = coefficients.astype(DTYPE_HIGH_PREC)
        poly_high_prec = MultivarPolynomial(coefficients, exponents, validate_input=False)

        for coeff_ctr in range(NR_COEFF_CHANGES):

            # evaluation @ uniformly random point
            x = ((np.random.rand(dim) - 0.5) * (2 * MAX_COEFF_MAGNITUDE))
            x = x.astype(FLOAT_DTYPE)

            # simply change coefficients of the found factorisation (cheap)
            coefficients = ((np.random.rand(nr_monomials, 1) - 0.5) * (2 * MAX_COEFF_MAGNITUDE))
            coefficients = coefficients.astype(FLOAT_DTYPE)

            # is testing for in_place=True at the same time
            poly_horner.change_coefficients(coefficients, validate_input=True, in_place=True)
            p_x_horner = poly_horner.eval(x)

            poly.change_coefficients(coefficients, validate_input=True, in_place=True)
            p_x = poly.eval(x)

            # in order to compare to a numerically accurate ground truth
            # increase the accuracy:
            # get info of data type supported by hardware: np.finfo(np.longdouble)
            # NOTE: the initial accuracy of the coefficients is the default float 64-bit accuracy
            # one must NOT create random 128-bit coefficients as ground truth
            # this would cause additional numerical error, because the algorithms have only 64-bit accuracy!
            # -> create 64-bit coefficients, then convert them to 128-bit!
            coefficients = coefficients.astype(DTYPE_HIGH_PREC)
            # p_x_expected = np.sum(coefficients)  # ground truth

            # ground truth at random point:
            # NOTE: need to deactivate jit compilation of naive approach
            # use higher prevision
            # suppress input validation
            poly_high_prec.change_coefficients(coefficients, validate_input=False, in_place=True)
            x = x.astype(DTYPE_HIGH_PREC)
            p_x_expected = poly_high_prec.eval(x)

            result = (poly, poly_horner, p_x_expected, p_x, p_x_horner)
            results.append(result)
            abs_numerical_error = abs(p_x_horner - p_x_expected)
            max_error = max(max_error, abs_numerical_error)
            sys.stdout.write(f'(poly #{poly_ctr + 1} coeff #{coeff_ctr + 1}, {(ctr_total + 1) / ctr_total_max:.1%})'
                             f' max numerical error: {max_error:.2e}\r')
            # sys.stdout.flush()
            if max_error > MAX_NUMERICAL_ERROR:
                # # DEBUG:
                # with open('coefficients.pickle', 'wb') as f:
                #     pickle.dump(coefficients, f)
                # with open('exponents.pickle', 'wb') as f:
                #     pickle.dump(exponents, f)
                raise AssertionError(f'numerical error {max_error:.2e} exceeded limit of {MAX_NUMERICAL_ERROR :.2e} ')
            ctr_total += 1

        sys.stdout.write("\n")  # move the cursor to the next line

    return results


class NumericalTest(unittest.TestCase):

    def test_numerical_stability(self):
        print('\nevaluating the numerical error:')
        results = []
        for dim, max_degree in product(DIM_RANGE, DEGREE_RANGE):
            results += evaluate_numerical_error(dim, max_degree)  # do not append list as entry

        with open(TEST_RESULTS_PICKLE, 'wb') as f:
            print(f'exporting numerical test results in {TEST_RESULTS_PICKLE}')
            pickle.dump(results, f)

        print('done.\n')



# -*- coding:utf-8 -*-

from setuptools import setup

setup(
    name='multivar_horner',
    packages=['multivar_horner'],
    description='python package implementing a multivariate Horner scheme for efficiently '
                'evaluating multivariate polynomials',
    # version: in VERSION file https://packaging.python.org/guides/single-sourcing-package-version/
    # With this approach you must make sure that the VERSION file is included in all your source
    # and binary distributions (e.g. add include VERSION to your MANIFEST.in).
    author='J. Michelfeit',
    author_email='python@michelfe.it',
    license='MIT licence',
    url='https://github.com/MrMinimal64/multivar_horner',  # use the URL to the github repo
    project_urls={
        "Source Code": "https://github.com/MrMinimal64/multivar_horner",
        "Documentation": "https://multivar_horner.readthedocs.io/en/latest/",
        "Changelog": "https://github.com/MrMinimal64/multivar_horner/blob/master/CHANGELOG.rst",
    },
    keywords='math mathematics polynomial polynomials polynomial-evaluation multivariate multivariate-polynomials'
             ' horner horner-scheme factorisation factorization',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    long_description='python package implementing a multivariate horner scheme '
                     'for efficiently evaluating multivariate polynomials \n\n'
                     'for more refer to the documentation: https://multivar_horner.readthedocs.io/en/latest/',
    install_requires=[
        'numpy>=1.16',
        'numba>=0.48',
    ],
    python_requires='>=3.6',
    # TODO
    # extras_require={'pytorch': [...]},
)



import os
import pickle
import time
import timeit
from pathlib import Path

import numpy as np

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib.patches import Rectangle
from multivar_horner.multivar_horner import HornerMultivarPolynomial, MultivarPolynomial
from scipy.special import binom
from tests.test_helpers import rnd_input_list, rnd_settings_list
from tests.test_settings import (
    DEGREE_RANGE, DIM_RANGE, EXPORT_RESOLUTION, NR_SAMPLES_SPEED_TEST,
    PLOTTING_DIR, SHOW_PLOTS, SPEED_RUN_PICKLE, TEST_RESULTS_PICKLE,
)


def get_plot_name(file_name='plot'):
    file_name = file_name.replace(' ', '_')
    Path(PLOTTING_DIR).mkdir(parents=True, exist_ok=True)
    return os.path.abspath(os.path.join(PLOTTING_DIR, file_name + '_' + str(time.time())[:-7] + '.png'))


def export_plot(fig, plot_title):
    # fig.set_size_inches(EXPORT_SIZE_X, EXPORT_SIZE_Y, forward=True)
    path = get_plot_name(plot_title)
    print(f'saving plot in {path}')
    plt.savefig(path, dpi=EXPORT_RESOLUTION)
    plt.clf()  # clear last figure


poly_settings_list = []
input_list = []
poly_class_instances = []


def setup_time_fct(poly_class):
    # store instances globally to directly use them for eval time test
    global poly_settings_list, poly_class_instances

    poly_class_instances = []
    for settings in poly_settings_list:
        poly_class_instances.append(poly_class(*settings))


def eval_time_fct():
    global poly_class_instances, input_list

    for instance, input in zip(poly_class_instances, input_list):
        instance.eval(input)


def test_setup_time_naive():
    setup_time_fct(MultivarPolynomial)


def test_setup_time_horner():
    setup_time_fct(HornerMultivarPolynomial)


def get_avg_num_ops():
    global poly_class_instances

    num_ops = 0
    for instance in poly_class_instances:
        num_ops += instance.get_num_ops()

    avg_num_ops = round(num_ops / len(poly_class_instances))
    return avg_num_ops


def time_preprocess(time, time_fmt='{:.2e}'):
    # valid_digits = 4
    # zero_digits = abs(min(0, int(log10(time))))
    # digits_to_print = zero_digits + valid_digits
    return time_fmt.format(time)


def difference(quantity1, quantity):
    speedup = round((quantity / quantity1 - 1), 1)
    if speedup < 0:
        speedup = round((quantity1 / quantity - 1), 1)
        if speedup > 10.0:
            speedup = round(speedup)

        return str(speedup) + ' x less'
    else:
        if speedup > 10.0:
            speedup = round(speedup)
        return str(speedup) + ' x more'


def compute_lucrativity(setup_horner, setup_naive, eval_horner, eval_naive):
    benefit_eval = eval_naive - eval_horner
    loss_setup = setup_horner - setup_naive
    # x * benefit = loss
    return round(loss_setup / benefit_eval)


attr_name_setup_time_naive = 'setup_time_naive'
attr_name_eval_time_naive = 'eval_time_naive'
attr_name_num_ops_naive = 'num_ops_naive'
attr_name_setup_time_horner = 'setup_time_horner'
attr_name_eval_time_horner = 'eval_time_horner'
attr_name_num_ops_horner = 'num_ops_horner'
attr_name_num_coeff = 'num_coeffs'


def speed_test_run(dim, degree, nr_samples, template):
    global poly_settings_list, input_list, poly_class_instances

    # TODO take times

    poly_settings_list = rnd_settings_list(nr_samples, dim, degree)
    input_list = rnd_input_list(nr_samples, dim, max_abs_val=1.0)

    setup_time_naive = timeit.timeit("test_setup_time_naive()", globals=globals(), number=1)
    setup_time_naive = setup_time_naive / NR_SAMPLES_SPEED_TEST  # = avg. per sample

    # poly_class_instances is not populated with the naive polynomial class instances
    # print(poly_class_instances[0])
    num_ops_naive = get_avg_num_ops()

    eval_time_naive = timeit.timeit("eval_time_fct()", globals=globals(), number=1)
    eval_time_naive = eval_time_naive / NR_SAMPLES_SPEED_TEST  # = avg. per sample

    setup_time_horner = timeit.timeit("test_setup_time_horner()", globals=globals(), number=1)
    setup_time_horner = setup_time_horner / NR_SAMPLES_SPEED_TEST  # = avg. per sample

    # poly_class_instances is not populated with the horner polynomial class instances
    # print(poly_class_instances[0])
    num_ops_horner = get_avg_num_ops()

    eval_time_horner = timeit.timeit("eval_time_fct()", globals=globals(), number=1)
    eval_time_horner = eval_time_horner / NR_SAMPLES_SPEED_TEST  # = avg. per sample

    setup_delta = difference(setup_time_naive, setup_time_horner)
    eval_delta = difference(eval_time_naive, eval_time_horner)
    ops_delta = difference(num_ops_naive, num_ops_horner)
    lucrative_after = compute_lucrativity(setup_time_horner, setup_time_naive, eval_time_horner, eval_time_naive)
    if lucrative_after < 0:
        lucrative_after = '-'  # never lucrative

    entries = [str(dim), str(degree), time_preprocess(setup_time_naive),
               time_preprocess(setup_time_horner), str(setup_delta),
               time_preprocess(eval_time_naive), time_preprocess(eval_time_horner),
               str(eval_delta), str(num_ops_naive), str(num_ops_horner), ops_delta, str(lucrative_after)]
    print(template.format(*entries))

    result = {
        attr_name_setup_time_naive: setup_time_naive,
        attr_name_eval_time_naive: eval_time_naive,
        attr_name_num_ops_naive: num_ops_naive,
        attr_name_setup_time_horner: setup_time_horner,
        attr_name_eval_time_horner: eval_time_horner,
        attr_name_num_ops_horner: num_ops_horner,
        # attr_name_num_coeff: num_coeffs, # TODO
    }
    return result


# TODO
def run_speed_benchmark():
    if os.path.isfile(SPEED_RUN_PICKLE):
        print(f'result file {SPEED_RUN_PICKLE} is already present. skipping data generation.')
        return

    # this also fulfills the purpose of testing the robustness of the code
    # (many random polygons are being created and evaluated)

    print('\nSpeed test:')
    print('testing {} evenly distributed random polynomials'.format(NR_SAMPLES_SPEED_TEST))
    print('average timings per polynomial:\n')

    print(' {0:11s}  |  {1:38s} |  {2:35s} |  {3:35s} | {4:20s}'.format('parameters', 'setup time (s)',
                                                                        'eval time (s)',
                                                                        '# operations', 'lucrative after '))
    template = '{0:3s} | {1:7s} | {2:10s} | {3:10s} | {4:13s} | {5:10s} | {6:10s} | {7:10s} | {8:10s} | ' \
               '{9:10s} | {10:10s} | {11:10s}'

    print(template.format('dim', 'max_deg', 'canonical', 'horner', 'delta', 'canonical',
                          'horner', 'delta', 'canonical', 'horner', 'delta', '   # evals'))
    print('=' * 160)

    all_results = []

    for dim in DIM_RANGE:
        dim_run_results = []
        for maximal_degree in DEGREE_RANGE:
            result_single_run = speed_test_run(dim, maximal_degree, NR_SAMPLES_SPEED_TEST, template)
            dim_run_results.append(result_single_run)

        print()  # empty line
        # dim_run_results = list(zip(dim_run_results))
        all_results.append(dim_run_results)

    print('writing results to file:', SPEED_RUN_PICKLE)
    with open(SPEED_RUN_PICKLE, 'wb') as f:
        pickle.dump(all_results, f)
    print('...done.\n')


def plot_speed_results():
    def extract_data(results, data_idx):
        data = []
        for dim_run_res in results:
            y_naive = []
            y_horner = []
            for degree_run_res in dim_run_res:
                y_naive.append(degree_run_res[0][data_idx])
                y_horner.append(degree_run_res[1][data_idx])

            data.append((y_naive, y_horner))

        return data

    def extract_data_diff(results, data_idx):
        data = []
        for dim_run_res in results:
            y_diff = []
            for degree_run_res in dim_run_res:
                y1 = degree_run_res[0][data_idx]
                y2 = degree_run_res[1][data_idx]
                y_diff.append(y1 - y2)

            data.append(y_diff)

        return data

    def extract_data_abs_horner(results, data_idx):
        data = []
        for dim_run_res in results:
            y_abs_horner = []
            for degree_run_res in dim_run_res:
                y = degree_run_res[1][data_idx]
                y_abs_horner.append(y)

            data.append(y_abs_horner)

        return data

    print('importing polynomial from file "{}"'.format(SPEED_RUN_PICKLE))
    with open(SPEED_RUN_PICKLE, 'rb') as f:
        all_results = pickle.load(f)

    print('plotting now...')

    # labels = ['avg. setup time increase / s', 'avg. evaluation time reduction / s', 'avg. #operations reduction']
    labels = ['time [s]', 'time [s]', 'avg. #operations reduction']
    file_names = ['setup_time_increase', 'eval_time_decrease', 'num_ops_decrease']

    # equal "spaced" colors
    # color_idx = np.linspace(0, 1, MAX_DIMENSION)
    # cm = plt.cm.gist_rainbow
    # use_logarithmic = [True, True, False]

    for run_idx in range(3):
        print(run_idx, file_names[run_idx])

        label = labels[run_idx]

        fig, ax = plt.subplots()

        obj_handles = []
        extra = Rectangle((0, 0), 1, 1, fc="w", fill=False, edgecolor='none', linewidth=0)
        obj_handles.append(extra)

        # data = extract_data(all_results, run_idx)
        data = extract_data_diff(all_results, run_idx)
        plt.plot([], [], ' ', label="dimension")
        for dim, dim_run_data in zip(reversed(DIM_RANGE), reversed(data)):
            # y_naive, y_horner = dim_run_data

            # plt.plot(x, y, color=c, alpha=alpha, **kwargs)
            # c = cm(color_idx[dim - 1])
            # TODO own plot setting for each, operations save log...
            plt.plot(DEGREE_RANGE, dim_run_data, 'x:',
                     # color=c,
                     label=str(dim), linewidth=3, markersize=15,
                     markeredgewidth=2.5)
            # plt.semilogy(DEGREE_RANGE, dim_run_data, 'x:', color=c, label=str(dim), linewidth=3, markersize=15,
            #              markeredgewidth=2.5)
            # plt.semilogy(DEGREE_RANGE, y_horner, color=c)
            # plt.semilogy(DEGREE_RANGE, y_naive, 'o--', color=c)

        # plt.semilogy(t, np.exp(-t / 5.0))
        plt.xticks(DEGREE_RANGE)
        plt.xlabel('polynomial maximal_degree')
        plt.ylabel(label)
        # plt.title(label)
        plt.legend()
        plt.grid(True)

        export_plot(fig, file_names[run_idx])
        if SHOW_PLOTS:
            plt.show()

    # TODO plot relative time improvement, but relative to naive not horner!

    print('...done.')


def extract_numerical_error_horner(result):
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    abs_error = abs(p_x_horner - p_x_expected)
    rel_error = abs_error / abs(p_x_expected)
    # return abs_error
    return rel_error


def extract_numerical_error_naive(result):
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    abs_error = abs(p_x - p_x_expected)
    rel_error = abs_error / abs(p_x_expected)
    # return abs_error
    return rel_error


def has_nonzero_err_horner(result):
    return extract_numerical_error_horner(result) != 0.0


def has_nonzero_err_naive(result):
    return extract_numerical_error_naive(result) != 0.0


def extract_numerical_difference(result):
    err_naive = extract_numerical_error_naive(result)
    err_horner = extract_numerical_error_horner(result)
    return err_horner - err_naive


def extract_poly_properties(result):
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    return poly.dim, poly.total_degree, poly.num_ops, poly_horner.num_ops


def extract_dimensionality(result):
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    return poly.dim


def extract_total_degree(result):
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    return poly.total_degree


def extract_maximal_degree(result):
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    return poly.maximal_degree


def extract_num_ops_naive(result):
    poly_dim, poly_degree, num_ops_naive, num_ops_horner = extract_poly_properties(result)
    return num_ops_naive


def extract_num_ops_horner(result):
    poly_dim, poly_degree, num_ops_naive, num_ops_horner = extract_poly_properties(result)
    return num_ops_horner


def extract_num_ops_benefit(result):
    poly_dim, poly_degree, num_ops_naive, num_ops_horner = extract_poly_properties(result)
    return num_ops_horner - num_ops_naive


def has_nonzero_num_ops_horner(result):
    return extract_num_ops_horner(result) != 0


def extract_nr_coeffs(result):  # identical in both polynomial representations!
    poly, poly_horner, p_x_expected, p_x, p_x_horner = result
    return poly_horner.num_monomials


def extract_sparsity(result):
    nr_coeffs = extract_nr_coeffs(result)
    dim = extract_dimensionality(result)
    deg = extract_total_degree(result)
    nr_coeffs_max = binom(dim + deg, deg)
    return nr_coeffs / nr_coeffs_max


def remove_zeros(l):
    return [x if x != 0.0 else None for x in l]


def rmv_zeros(*args):
    out = []
    for entries in zip(*args):
        if 0.0 not in entries:
            out.append(entries)

    return list(zip(*out))


def average_discrete(l1, l2):
    """
    average the entries of l1 where l2 is identical
    """
    assert len(l1) == len(l2)
    l1 = np.array(l1)
    l2 = np.array(l2)
    unique_entries = np.unique(l2)
    avgs = np.array([np.average(l1[l2 == e]) for e in unique_entries])
    assert len(avgs) == len(unique_entries)
    return avgs, unique_entries


def make_relative(a, b):
    return b / a


# def compute_log(entry):
#     if entry is None:
#         return None
#     return log10(entry)
#
#
# def convert2log(l):
#     return list(map(compute_log, l))


def plot_num_err_heatmap(results):
    dims = list(map(extract_dimensionality, results))
    # degs = list(map(extract_total_degree, results))
    maximal_degs = list(map(extract_maximal_degree, results))
    numerical_err_naive = list(map(extract_numerical_error_naive, results))
    numerical_err_horner = list(map(extract_numerical_error_horner, results))

    attr_name_numerical_err_naive = 'numerical error naive'
    attr_name_numerical_err_horner = 'numerical error horner'
    attr_name_numerical_err_rel = 'relative numerical error'
    attr_name_dim = 'dimension'
    attr_name_deg = 'maximal degree'

    df = pd.DataFrame({
        attr_name_numerical_err_naive: numerical_err_naive,
        attr_name_numerical_err_horner: numerical_err_horner,
        attr_name_deg: maximal_degs,
        attr_name_dim: dims,
    })

    df_avg = pd.DataFrame()

    # average for dim and degree
    for dim in DIM_RANGE:
        for maximal_deg in DEGREE_RANGE:
            selection_idxs = (df[attr_name_dim] == dim) & (df[attr_name_deg] == maximal_deg)
            if not selection_idxs.any():
                continue
            df_selected = df[selection_idxs]
            avg_num_err_naive = df_selected[attr_name_numerical_err_naive].mean()
            avg_num_err_horner = df_selected[attr_name_numerical_err_horner].mean()
            if avg_num_err_naive == avg_num_err_horner:
                avg_num_err_naive_rel = 1.0
            else:
                avg_num_err_naive_rel = avg_num_err_naive / avg_num_err_horner

            entry = pd.Series({
                attr_name_numerical_err_rel: avg_num_err_naive_rel,
                attr_name_numerical_err_naive: avg_num_err_naive,
                attr_name_numerical_err_horner: avg_num_err_horner,
                attr_name_deg: maximal_deg,
                attr_name_dim: dim,
            })
            df_avg = df_avg.append(entry, ignore_index=True)

    df_avg[attr_name_deg] = df_avg[attr_name_deg].astype(int, copy=False)
    df_avg[attr_name_dim] = df_avg[attr_name_dim].astype(int, copy=False)

    # sns.relplot(x=attr_name_dim, y=attr_name_deg, data=df_avg,
    #             hue=attr_name_numerical_err_rel,
    #             size=attr_name_numerical_err_rel,
    #             sizes=(100,300),
    #             )
    heatmap_data = df_avg.pivot(attr_name_deg, attr_name_dim, attr_name_numerical_err_rel)
    heatmap_data = heatmap_data.iloc[::-1]  # reverse

    sns.heatmap(heatmap_data, annot=True,
                # fmt="d",
                linewidths=3,
                # cmap="YlGnBu",
                )
    plot_title = 'avg. numerical error with canonical form relative to Horner factorisation'
    plt.title(plot_title)
    # plt.xlabel()
    # locs, labels = plt.xticks()  # Get locations and labels
    # plt.xticks(locs, labels=[int(x) for x in labels])  # Set locations and labels
    fig = plt.gcf()
    title = 'numerical error heatmap'
    export_plot(fig, plot_title=title)


def plot_num_error_growth_comparison(results):
    numerical_err_naive = list(map(extract_numerical_error_naive, results))
    numerical_err_horner = list(map(extract_numerical_error_horner, results))
    num_coeffs = list(map(extract_nr_coeffs, results))

    # average first! before removing zeros
    numerical_err_naive_avg, num_coeffs_unique_naive = average_discrete(numerical_err_naive, num_coeffs)
    numerical_err_horner_avg, num_coeffs_unique_horner = average_discrete(numerical_err_horner, num_coeffs)

    numerical_err_naive_avg, num_coeffs_unique_naive = rmv_zeros(numerical_err_naive_avg, num_coeffs_unique_naive)
    numerical_err_horner_avg, num_coeffs_unique_horner = rmv_zeros(numerical_err_horner_avg, num_coeffs_unique_horner)

    attr_name_representation = 'representation'
    attr_name_num_coeff = 'number of coefficients'
    attr_name_numerical_err = 'average relative numerical error'
    repr_name_horner = 'Horner factorisation'
    repr_name_naive = 'canonical form'

    df = pd.DataFrame()

    df_naive = pd.DataFrame({attr_name_num_coeff: num_coeffs_unique_naive,
                             attr_name_numerical_err: numerical_err_naive_avg,
                             attr_name_representation: repr_name_naive})
    df_naive[attr_name_representation] = repr_name_naive

    df_horner = pd.DataFrame({attr_name_num_coeff: num_coeffs_unique_horner,
                              attr_name_numerical_err: numerical_err_horner_avg, })
    df_horner[attr_name_representation] = repr_name_horner

    df = df.append(df_naive, ignore_index=True)
    df = df.append(df_horner, ignore_index=True)

    sns.scatterplot(x=attr_name_num_coeff, y=attr_name_numerical_err, hue=attr_name_representation, data=df,
                    # alpha=ALPHA
                    )
    # plot = sns.relplot(x=attr_name_num_coeff, y=attr_name_numerical_err, hue=attr_name_representation,
    #                    style=attr_name_representation,
    #                    kind="line", data=df,)
    # plt.legend()
    # plot.add_legend(bbox_to_anchor=(1.05, 0), loc=2, borderaxespad=0.)
    title = 'avg numerical error VS nr coeff'
    plt.xscale('log')
    plt.yscale('log')
    plt.grid(True)
    fig = plt.gcf()
    export_plot(fig, plot_title=title)


def filter_results(filter_fct, results):
    results_filtered = list(filter(filter_fct, results))
    print(f'filtering data: out of {len(results)} entries {len(results_filtered)} remain')
    return results_filtered


def plot_num_coeffs2num_ops(results):
    attr_name_num_coeff = 'number of coefficients'
    attr_name_num_ops = 'number of operations'
    attr_name_representation = 'representation'
    attr_name_dimensionality = 'dimensionality'

    repr_name_horner = 'Horner factorisation'
    repr_name_naive = 'canonical form'

    results = filter_results(has_nonzero_num_ops_horner, results)

    # compare the number of operations to the number of coefficients
    # color by dimensions
    dims = list(map(extract_dimensionality, results))
    num_ops_naive = list(map(extract_num_ops_naive, results))
    num_ops_horner = list(map(extract_num_ops_horner, results))
    num_coeffs = list(map(extract_nr_coeffs, results))

    df = pd.DataFrame()

    df_naive = pd.DataFrame({
        attr_name_num_coeff: num_coeffs,
        attr_name_num_ops: num_ops_naive,
        attr_name_dimensionality: dims,
    })
    df_naive[attr_name_representation] = repr_name_naive

    df_horner = pd.DataFrame({
        attr_name_num_coeff: num_coeffs,
        attr_name_num_ops: num_ops_horner,
        attr_name_dimensionality: dims,
    })
    df_horner[attr_name_representation] = repr_name_horner

    df = df.append(df_naive, ignore_index=True)
    df = df.append(df_horner, ignore_index=True)

    # compare the number of operations to the number of coefficients
    # num_ops_horner, num_coeffs1, dim1 = rmv_zeros(num_ops_horner, num_coeffs, dim)
    sns.scatterplot(x=attr_name_num_coeff, y=attr_name_num_ops, data=df,
                    # hue=attr_name_dimensionality,
                    hue=attr_name_representation,
                    )
    title = 'opsVS coeffs'
    # plot.set_title(title)
    plt.xscale('log')
    plt.yscale('log')
    plt.grid(True)
    fig = plt.gcf()
    export_plot(fig, plot_title=title)

    # c = cm(color_idx[dim - 1])
    sns.scatterplot(x=attr_name_num_coeff, y=attr_name_num_ops, hue=attr_name_dimensionality, data=df_naive)
    title = 'ops naive VS coeffs'
    # plot.set_title(title)
    plt.xscale('log')
    plt.yscale('log')
    plt.grid(True)
    # TODO ticks
    # r'$\mu=100,\ \sigma=15$'
    fig = plt.gcf()
    export_plot(fig, plot_title=title)

    # num_ops_horner, num_coeffs1, dim1 = rmv_zeros(num_ops_horner, num_coeffs, dim)
    sns.scatterplot(x=attr_name_num_coeff, y=attr_name_num_ops, hue=attr_name_dimensionality, data=df_horner)
    title = 'ops horner VS coeffs'
    plt.xscale('log')
    plt.yscale('log')
    plt.grid(True)
    fig = plt.gcf()
    export_plot(fig, plot_title=title)


def plot_numerical_error():
    print(f'generating numerical benchmark plots. reading data from {TEST_RESULTS_PICKLE}...')
    try:
        with open(TEST_RESULTS_PICKLE, 'rb') as f:
            results = pickle.load(f)
    except FileNotFoundError:
        print('no data of numerical errors found. run the numerical tests in test_it.py first')
        return

    # results_filtered_naive = filter_results(has_nonzero_err_naive, results)
    # results_filtered_horner = filter_results(has_nonzero_err_horner, results)

    plot_num_err_heatmap(results)
    plot_num_error_growth_comparison(results)
    plot_num_coeffs2num_ops(results)
    print('plotting numerical benchmarks done.')


if __name__ == '__main__':
    # TODO plot "lucrative after"

    sns.set_context("paper")
    # plt.rcParams.update({'font.size': 35})

    # TODO do not average results. measure time of a single polynomial.
    # TODO plot with seaborn in the same fashion as numerical test
    # TODO compare & plot the performance wrt. the "density" of the polynomials. sparse <-> fully occupied
    # TODO make num tested polynomials dependent on parameters (size, time) (faster runs are more noisy)
    # run_speed_benchmark()  # generate speed test data
    # TODO include plots in documentation
    # plot_speed_results()  # create plots with speed test data

    plot_numerical_error()  # create plots with numerical test data



# TODO
# for Ahead-Of-Time Compilation:
# from numba.pycc import CC
# cc = CC('compiled_helpers', )
# # Uncomment the following line to print out the compilation steps
# cc.verbose = True


# precompiled time critical helper functions
import numpy as np
from numba import b1, f8, jit, u4


# TODO define and import dtypes globally
# @jit(f8(f8[:], f8[:], u4[:, :]), nopython=True, cache=True)
def naive_eval(x, coefficients, exponents):
    nr_coeffs = len(coefficients)
    # nr_monomials,nr_dims = exponents.shape
    # assert nr_monomials == nr_coeffs
    # assert len(x) == nr_dims
    acc = 0.0
    for i in range(nr_coeffs):
        acc = acc + coefficients[i] * np.prod(np.power(x, exponents[i, :]))
    return acc

    # equivalent one liner:
    # TypingError: numba doesn't support kwarg for prod
    # return np.sum(coefficients.T * np.prod(np.power(x, exponents), axis=1), axis=1)[0]


# @cc.export('eval_compiled', 'f8(f8[:], f8[:], u4[:, :], u4[:, :], u4[:, :], u4[:, :], b1[:], u4)')
@jit(f8(f8[:], f8[:], u4[:, :], u4[:, :], u4[:, :], u4[:, :], b1[:], u4), nopython=True, cache=True, debug=True)
def eval_recipe(x, value_array, copy_recipe, scalar_recipe, monomial_recipe, tree_recipe, tree_ops, root_value_address):
    # IMPORTANT: the order of following the recipes is not arbitrary!
    #   scalar factors need to be evaluated before monomial factors depending on them...

    # in order to evaluate scalar factors with exponent 1, no exponentiation operation is required
    # simply copy the values of x to the value array
    # copy recipe instruction encoding: target, source
    # [target, source] = copy_recipe[i, :]
    for i in range(copy_recipe.shape[0]):
        # value_array[target] = x[source1]
        value_array[copy_recipe[i, 0]] = x[copy_recipe[i, 1]]

    # print('computing scalar factors: ...')
    # scalar recipe instruction encoding: target, source, exponent
    # [target, source1, exponent] = scalar_recipe[i, :]
    for i in range(scalar_recipe.shape[0]):
        # print('value[{}] = {} ^ {}'.format(target, x[source1], exponent))
        # value_array[target] = x[source1] ** exponent
        value_array[scalar_recipe[i, 0]] = x[scalar_recipe[i, 1]] ** scalar_recipe[i, 2]

    # # DEBUG:
    # accessed_idxs = set()

    # print('computing monomial factors: ...')
    # monomial recipe instruction encoding: target, source1, source2
    # [target, source1, source2] = monomial_recipe[i, :]
    for i in range(monomial_recipe.shape[0]):
        # print('value[{}] = {} * {} (idx: {}, {})'.format(target, value_array[source1], value_array[source2], source1,
        #                                                 source2))
        # value_array[target] = value_array[source1] * value_array[source2]
        value_array[monomial_recipe[i, 0]] = value_array[monomial_recipe[i, 1]] * value_array[monomial_recipe[i, 2]]

        # # DEBUG:
        # accessed_idxs.add(monomial_recipe[i, 1])
        # accessed_idxs.add(monomial_recipe[i, 2])

    # print('evaluating factorisation tree: ...')
    # tree recipe instruction encoding: target, source
    # [target, source] = tree_recipe[i, :]
    # separate operation array: *op_id*
    # value_array[target] = value_array[target] *op* value_array[source]
    for i in range(tree_recipe.shape[0]):
        target = tree_recipe[i, 0]
        source = tree_recipe[i, 1]
        if tree_ops[i]:  # ADDITION: 1
            # print('value[{}] = {} + {}'.format(target, value_array[target], value_array[source1]))
            # value_array[target] = value_array[target] + value_array[source1]
            value_array[target] = value_array[target] + value_array[source]
        else:  # MULTIPLICATION: 0
            # print('value[{}] = {} * {}'.format(target, value_array[target], value_array[source1]))
            # value_array[target] = value_array[target] * value_array[source1]
            value_array[target] = value_array[target] * value_array[source]

        # # DEBUG:
        # accessed_idxs.add(target)
        # accessed_idxs.add(source)

    # # DEBUG:
    # all_idxs = {x for x in range(len(value_array))}
    # non_accessed_idxs = all_idxs - accessed_idxs
    # # coefficients are stored in the indices until one before the root address
    # coefficient_idxs = {x for x in range(root_value_address)}
    # non_accessed_coefficients = non_accessed_idxs & coefficient_idxs
    # if len(non_accessed_coefficients) > 0:
    #     raise ValueError(f'BUG: these coefficients have been accessed: {non_accessed_coefficients}')
    #
    # # NOTE: no indices must be accessed when the polynomial is a constant
    # if len(non_accessed_idxs) > 0 and len(value_array) > 1:
    #     raise ValueError(f'BUG: these idxs have been accessed: {non_accessed_idxs}')

    return value_array[root_value_address]  # return value of the root node


@jit(u4(u4[:]), nopython=True, cache=True)
def num_ops_1D_horner(unique_exponents):
    """
    :param unique_exponents: np array of unique exponents sorted in increasing order without 0
    :return: the number of operations of the one dimensional Horner factorisation
        without counting additions (just MUL & POW) and without considering the coefficients
    do not count additions and exponentiations. just multiplications
    do not consider the multiplications with the coefficients (amount stays constant)


    NOTE: in 1D the Horner factorisation is both unique and optimal (minimal amount of operations)
    -> gives a lower bound for the amount of required operations
    (=required property for usage as cost estimation heuristic)
    """
    nr_unique_exponents = unique_exponents.shape[0]
    # assert not np.any(unique_exponents == 0), "the exponent 0 must not be present"
    if nr_unique_exponents == 0:
        return 0

    # one MUL operation is required !between! all factors in the factorisation chain
    # the amount of factors (= "length of factorisation chain") is equal to the amount of unique existing exponents
    num_ops = - 1

    # start with exponent 0 (not in unique exponents)
    # the difference between one and the next exponent determines if a POW operation is needed to evaluate a factor
    # unique exponents are sorted: prev_exp < exp
    prev_exp = 0
    for i in range(nr_unique_exponents):
        exp = unique_exponents[i]
        # the exponents MUST increase -> count at least one operation (multiplication)
        # for every exponent difference >1
        # count exponent-1 additional multiplications for computing the exponentiations
        exp_diff = exp - prev_exp
        num_ops += exp_diff
        prev_exp = exp

    return num_ops


@jit(u4(u4[:, :]), nopython=True, cache=True)
def count_num_ops(exponent_matrix):
    """ counts the amount of multiplications required during evaluation

    under the assumption: this polynomial representation does not get factorised any further
    do not count additions and exponentiations

    1 multiplication required for every coefficient when there is a monomial (non zero exponent) present
    one multiplication for multiplying the scalar factors with each other
    (amount of non zero exponents in each monomial -1)
    exponent - 1 multiplications for every evaluation of an exponentiation of every scalar factor
    = the total sum of exponents
    """
    return np.sum(exponent_matrix)


@jit(u4(u4, u4), nopython=True, cache=True)
def factor_num_ops(dim, exp):
    """
    NOTE: all factors are scalars: x^i
    count every exponentiation as exponent-1 multiplications

    :return: the amount of operations required to evaluate the scalar factor
    """
    return exp - 1


@jit(u4[:](u4, u4[:], u4[:], u4[:, :]), nopython=True, cache=True)
def compile_usage(dim, usage_vector, unique_exponents, exponent_matrix):
    """
    :return: a vector with the usage count of every unique exponent
    """

    for i in range(exponent_matrix.shape[0]):
        exp = exponent_matrix[i, dim]
        for j in range(len(unique_exponents)):
            if exp < unique_exponents[j]:
                break
            usage_vector[j] += 1

    return usage_vector


@jit(b1[:](u4, b1[:], u4[:], u4[:], u4[:, :]), nopython=True, cache=True)
def compile_valid_options(dim, valid_option_vector, usage_vector, unique_exponents, exponent_matrix):
    if len(valid_option_vector) == 0:
        # there are no unique exponents
        return valid_option_vector

    usage_vector = compile_usage(dim, usage_vector, unique_exponents, exponent_matrix)

    for exp_idx in range(usage_vector.size):
        # stop at the highest exponent having a usage >=2
        if usage_vector[exp_idx] >= 2:
            valid_option_vector[exp_idx] = True
        else:
            # all higher exponents have a lower usage!
            break

    return valid_option_vector


@jit(u4(u4, u4, u4[:, :]), nopython=True, cache=True)
def count_usage(dim, exp, exponent_matrix):
    """
    :return: the amount of times a scalar factor appears in the monomials
    """

    usage_cnt = 0
    for i in range(exponent_matrix.shape[0]):
        if exponent_matrix[i, dim] >= exp:
            usage_cnt += 1

    return usage_cnt



# -*- coding:utf-8 -*-

import pickle
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List

import numpy as np

from multivar_horner.factorisation_classes import HeuristicFactorisationRoot, OptimalFactorisationRoot
from multivar_horner.global_settings import (
    BOOL_DTYPE, DEBUG, DEFAULT_PICKLE_FILE_NAME, FLOAT_DTYPE, TYPE_1D_FLOAT, TYPE_2D_INT, UINT_DTYPE,
)
from multivar_horner.helper_classes import FactorContainer
from multivar_horner.helper_fcts import (
    rectify_coefficients, rectify_construction_parameters, rectify_query_point,
    validate_coefficients, validate_construction_parameters, validate_query_point,
)
from multivar_horner.helpers_fcts_numba import count_num_ops, eval_recipe, naive_eval


# is not a helper function to make it an importable part of the package
def load_pickle(path: str = DEFAULT_PICKLE_FILE_NAME) -> 'AbstractPolynomial':
    print('importing polynomial from file "{}"'.format(path))
    with open(path, 'rb') as f:
        return pickle.load(f)


'''
    TODO
    POSSIBLE IMPROVEMENTS:

    MultivarPoly:

    - also make use of the concept of 'recipes' for efficiently evaluating the polynomial
        skipping the most unnecessary operations
    - add option to skip this optimisation

    HornerMultivarPoly:


    - optimise factor evaluation (save instructions, 'factor factorisation'):
    a monomial factor consists of scalar factors and in turn some monomial factors consist of other monomial factors
    -> the result of evaluating a factor can be reused for evaluating other factors containing it
    -> find the optimal 'factorisation' of the factors themselves
    -> set the factorisation_idxs of each factor in total_degree to link the evaluation appropriately
    idea:
        choose  'Goedel IDs' as the monomial factor ids
        then the id of a monomial is the product of the ids of its scalar factors
        find the highest possible divisor among all factor ids
        (corresponds to the 'largest' factor included in the monomial)
        this leads to a minimal factorisation for evaluating the monomial values quickly

    - add option to skip this optimisation to save build time

    - optimise space requirement:
     after building a factorisation tree for the factors themselves,
     then use its structure to cleverly reuse storage space
     -> use compiler construction theory: minimal assembler register assignment, 'graph coloring'...

    - optimise 'copy recipe': avoid copy operations for accessing values of x
        problem: inserting x into the value array causes operations as well and
            complicates address assigment and recipe compilation

    -  when the polynomial does not depend on all variables, build a wrapper to maintain the same "interface"
        but internally reduce the dimensionality, this reduced the size of the numpy arrays -> speed, storage benefit

    - the evaluation of subtrees is independent and could theoretically be done in parallel
        probably not worth the effort. more reasonable to just evaluate multiple polynomials in parallel
'''


# TODO properties: num ops...

class DummyClass:
    """ a class

    Attributes:
        a: test attribute
    """

    # FIXME: creates duplicate entries in Sphinx autodoc
    __slots__ = ['a']


# TODO docstring attributes in parent class, automatic inheritance in sphinx autodoc api docu
#   -> prevent duplicate docstring in MultivarPolynomial
class AbstractPolynomial(ABC):
    """ an abstract class for representing a multivariate polynomial
    """

    # prevent dynamic attribute assignment (-> safe memory)
    # FIXME: creates duplicate entries in Sphinx autodoc
    __slots__ = ['compute_representation', 'coefficients', 'euclidean_degree', 'exponents', 'num_monomials', 'num_ops',
                 'dim', 'maximal_degree', 'total_degree', 'unused_variables', 'representation']

    def __init__(self, coefficients: TYPE_1D_FLOAT, exponents: TYPE_2D_INT, rectify_input: bool = False,
                 validate_input: bool = False, compute_representation: bool = False):

        if rectify_input:
            coefficients, exponents = rectify_construction_parameters(coefficients, exponents)

        if validate_input:
            validate_construction_parameters(coefficients, exponents)

        self.coefficients: np.ndarray = coefficients
        self.exponents: np.ndarray = exponents
        self.compute_representation = compute_representation

        self.num_monomials: int = self.exponents.shape[0]
        self.dim: int = self.exponents.shape[1]
        self.unused_variables = np.where(~np.any(self.exponents.astype(BOOL_DTYPE), axis=1))[0]
        self.total_degree: int = np.max(np.sum(self.exponents, axis=0))
        self.euclidean_degree: float = np.max(np.linalg.norm(self.exponents, ord=2, axis=0))
        self.maximal_degree: int = np.max(self.exponents)

        self.num_ops: int = 0
        self.representation: str = 'p(x)'

    def __str__(self):
        return self.representation

    def __repr__(self):
        return self.representation

    def __call__(self, *args, **kwargs) -> float:
        return self.eval(*args, **kwargs)

    def get_num_ops(self) -> int:
        return self.num_ops

    def compute_string_representation(self, *args, **kwargs) -> str:
        """ computes a string representation of the polynomial and sets self.representation

        Returns:
            a string representing this polynomial instance
        """
        # self.representation has already been set during construction
        return self.representation

    def export_pickle(self, path: str = DEFAULT_PICKLE_FILE_NAME):
        print('storing polynomial in file "{}" ...'.format(path))
        with open(path, 'wb') as f:
            pickle.dump(self, f)
        print('...done.\n')

    # TODO test
    def get_partial_derivative(self, i: int, *args, **kwargs) -> 'AbstractPolynomial':
        """ retrieves a partial derivative

        Note:
            all given additional arguments will be passed to the constructor of the derivative polynomial

        Args:
            i: dimension to derive with respect to.
                ATTENTION: dimension counting starts with 1 (i >= 1)

        Returns:
            the partial derivative of this polynomial wrt. the i-th dimension
        """

        assert (0 < i <= self.dim), 'invalid dimension i given'
        coord_index = i - 1
        # IMPORTANT: do not modify the stored coefficient and exponent arrays of self!
        # set all the coefficients not depending on the i-th coordinate to 0
        # this simply means not adding them to the list of coefficients of the new polynomial class instance
        active_rows = np.where(self.exponents[:, coord_index] >= 1)[0]

        new_coefficients = self.coefficients[active_rows]
        new_exponents = self.exponents[active_rows, :]

        if DEBUG:  # TODO move to tests
            assert new_coefficients.shape[0] == new_exponents.shape[0]
            assert new_coefficients.shape[1] == 1 and len(new_coefficients.shape) == 2
            assert new_exponents.shape[1] == self.dim and len(new_exponents.shape) == 2

        # multiply the coefficients with the exponent of the i-th coordinate
        # f(x) = a x^b
        # f'(x) = ab x^(b-1)
        new_coefficients = np.multiply(new_coefficients.flatten(), new_exponents[:, coord_index])
        new_coefficients = new_coefficients.reshape(-1, 1)

        # reduce the the exponent of the i-th coordinate by 1
        new_exponents[:, coord_index] -= 1

        # must call the proper constructor method also for inherited classes
        return self.__class__(new_coefficients, new_exponents, *args, **kwargs)

    def get_gradient(self, *args, **kwargs) -> List['AbstractPolynomial']:
        """
        Note:
            all arguments will be passed to the constructor of the derivative polynomials

        Returns:
             the list of all partial derivatives
        """
        return [self.get_partial_derivative(i, *args, **kwargs) for i in range(1, self.dim + 1)]

    def change_coefficients(self, coefficients: TYPE_1D_FLOAT, rectify_input: bool = False,
                            validate_input: bool = False, compute_representation: bool = False, in_place: bool = False,
                            *args, **kwargs) -> 'AbstractPolynomial':

        if rectify_input:
            coefficients = rectify_coefficients(coefficients)

        if validate_input:
            validate_coefficients(coefficients)

        assert len(coefficients) == self.num_monomials

        if in_place:
            updated_poly = self
        else:
            updated_poly = deepcopy(self)

        updated_poly.compute_representation = compute_representation

        updated_poly.coefficients = coefficients
        updated_poly.compute_string_representation(*args, **kwargs)
        return updated_poly

    @abstractmethod
    def eval(self, x: TYPE_1D_FLOAT, validate_input=True) -> float:
        pass


class MultivarPolynomial(AbstractPolynomial):
    """ a representation of a multivariate polynomial in 'canonical form' (without any factorisation)

    Args:
        coefficients: ndarray of floats with shape (N,1)
            representing the coefficients of the monomials
            NOTE: coefficients with value 0 and 1 are allowed and will not affect the internal representation,
            because coefficients must be replaceable
        exponents: ndarray of unsigned integers with shape (N,m)
            representing the exponents of the monomials
            where m is the number of dimensions (self.dim),
            the ordering corresponds to the ordering of the coefficients, every exponent row has to be unique!
        rectify_input: bool, default=False
            whether to convert coefficients and exponents into compatible numpy arrays
            with this set to True, coefficients and exponents can be given in standard python arrays
        validate_input: bool, default=False
            whether to check if coefficients and exponents fulfill the requirements (shape, data type etc.)
        compute_representation: bool, default=False
            whether to compute a string representation of the polynomial

    Attributes:
        num_monomials: the amount of coefficients/monomials N of the polynomial
        dim: the dimensionality m of the polynomial
            NOTE: the polynomial needs not to actually depend on all m dimensions
        unused_variables: the dimensions the polynomial does not depend on
        num_ops: the amount of mathematical operations required to evaluate the polynomial in this representation
        representation: a human readable string visualising the polynomial representation

    Raises:
        TypeError: if ``validate_input=True`` and coefficients or exponents are not given as ndarrays
            of the required dtype
        ValueError: if ``validate_input=True`` and coefficients or exponents do not have the required shape or
            do not fulfill the other requirements or ``rectify_input=True`` and there are negative exponents
    """

    def __init__(self, coefficients: TYPE_1D_FLOAT, exponents: TYPE_2D_INT, rectify_input: bool = False,
                 validate_input: bool = False, compute_representation: bool = False, *args, **kwargs):

        super(MultivarPolynomial, self).__init__(coefficients, exponents, rectify_input, validate_input,
                                                 compute_representation)

        # NOTE: count the number of multiplications of the representation
        # not the actual amount of operations required by the naive evaluation with numpy arrays
        self.num_ops = count_num_ops(self.exponents)
        self.compute_string_representation(*args, **kwargs)

    def compute_string_representation(self, coeff_fmt_str: str = '{:.2}', factor_fmt_str: str = 'x_{dim}^{exp}', *args,
                                      **kwargs) -> str:
        representation = '[#ops={}] p(x)'.format(self.num_ops)
        if self.compute_representation:
            representation += ' = '
            monomials = []
            for i, exp_vect in enumerate(self.exponents):
                monomial = [coeff_fmt_str.format(self.coefficients[i, 0])]
                for dim, exp in enumerate(exp_vect):
                    # show all operations, even 1 * x_i^0
                    monomial.append(factor_fmt_str.format(**{'dim': dim + 1, 'exp': exp}))

                monomials.append(' '.join(monomial))

            representation += ' + '.join(monomials)

        self.representation = representation
        return self.representation

    def eval(self, x: TYPE_1D_FLOAT, rectify_input: bool = False, validate_input: bool = False, ) -> float:
        """ computes the value of the polynomial at query point x

        makes use of fast ``Numba`` just in time compiled functions

        Args:
            x: ndarray of floats with shape = [self.dim] representing the query point
            rectify_input: bool, default=False
                whether to convert coefficients and exponents into compatible numpy arrays
                with this set to True, the query point x can be given in standard python arrays
            validate_input: check if the query point x fulfills the requirements

        Returns:
             the value of the polynomial at point x

        Raises:
            TypeError: if ``validate_input=True`` and x is not given as ndarray of dtype float
            ValueError: if ``validate_input=True`` and x does not have the shape ``[self.dim]``
        """

        if rectify_input:
            x = rectify_query_point(x)
        if validate_input:
            validate_query_point(x)
        if x.shape[0] != self.dim:
            raise ValueError(f'the query point x does not have the required dimensionality {self.dim}')

        return naive_eval(x, self.coefficients.flatten(), self.exponents)


class HornerMultivarPolynomial(AbstractPolynomial):
    """ a representation of a multivariate polynomial using Horner factorisation

    the polynomial is being evaluated by fast just in time compiled functions
    using precompiled "recipes" of instructions.

    Args:
        coefficients: ndarray of floats with shape (N,1)
            representing the coefficients of the monomials
            NOTE: coefficients with value 0 and 1 are allowed and will not affect the internal representation,
            because coefficients must be replaceable
        exponents: ndarray of unsigned integers with shape (N,m)
            representing the exponents of the monomials
            where m is the number of dimensions (self.dim),
            the ordering corresponds to the ordering of the coefficients, every exponent row has to be unique!
        rectify_input: bool, default=False
            whether to convert coefficients and exponents into compatible numpy arrays
            with this set to True, coefficients and exponents can be given in standard python arrays
        validate_input: bool, default=False
            whether to check if coefficients and exponents fulfill the requirements (shape, data type etc.)
        compute_representation: bool, default=False
            whether to compute a string representation of the polynomial

        keep_tree: whether the factorisation tree object should be kept in memory after finishing factorisation
        find_optimal: whether a search over all possible factorisations should be done in total_degree to find
            an optimal factorisation in the sense of a minimal amount required operations for evaluation


    Attributes:
        num_monomials: the amount of coefficients/monomials N of the polynomial
        dim: the dimensionality m of the polynomial
            NOTE: the polynomial needs not to actually depend on all m dimensions
        unused_variables: the dimensions the polynomial does not depend on
        num_ops: the amount of mathematical operations required to evaluate the polynomial in this representation
        representation: a human readable string visualising the polynomial representation

        total_degree: the usual notion of degree for a polynomial.
            = the maximum sum of exponents in any of its monomials
            = the maximum l_1-norm of the exponent vectors of all monomials
            in contrast to 1D polynomials, different concepts of degrees exist for polynomials in multiple dimensions.
            following the naming in [1] L. Trefethen, “Multivariate polynomial approximation in the hypercube”,
            Proceedings of the American Mathematical Society, vol. 145, no. 11, pp. 4837–4844, 2017.
        euclidean_degree: the maximum l_2-norm of the exponent vectors of all monomials.
            NOTE: this is not in general an integer
        maximal_degree: the largest exponent in any of its monomials
            = the maximum l_infinity-norm of the exponent vectors of all monomials

        factorisation_tree: the object oriented, recursive data structure representing the factorisation
            (only if keep_tree=True)
        factor_container: the object containing all (unique) factors of the factorisation (only if keep_tree=True)
        root_value_idx: the index in the value array where the value of this polynomial
            (= root of the factorisation_tree) will be stored
        value_array_length: the amount of addresses (storage) required to evaluate the polynomial.
            for evaluating the polynomial in tree form intermediary results have to be stored in a value array.
            the value array begins with the coefficients of the polynomial.
            (without further optimisation) every factor requires its own address.

        copy_recipe: ndarray encoding the operations required to evaluate all scalar factors with exponent 1
        scalar_recipe: ndarray encoding the operations required to evaluate all remaining scalar factors
        monomial_recipe: ndarray encoding the operations required to evaluate all monomial factors
        tree_recipe: ndarray encoding the addresses required to evaluate
            the polynomial values of the factorisation_tree.
        tree_ops: ndarray encoding the type of operation required to evaluate
            the polynomial values of the factorisation_tree.
            encoded as a boolean ndarray separate from tree_recipe,
            since only the two operations ADD & MUL need to be encoded.

    Raises:
        TypeError: if ``validate_input=True`` and coefficients or exponents are not given as ndarrays
            of the required dtype
        ValueError: if ``validate_input=True`` and coefficients or exponents do not have the required shape or
            do not fulfill the other requirements
    """
    # __slots__ declared in parents are available in child classes. However, child subclasses will get a __dict__
    # and __weakref__ unless they also define __slots__ (which should only contain names of any additional slots).
    # FIXME: creates duplicate entries in Sphinx autodoc
    __slots__ = ['copy_recipe', 'factorisation_tree', 'factor_container', 'monomial_recipe', 'root_value_idx',
                 'tree_recipe', 'tree_ops', 'scalar_recipe', 'value_array_length']

    def __init__(self, coefficients, exponents, rectify_input=False, validate_input=False, keep_tree=False,
                 compute_representation=False, find_optimal=False, *args, **kwargs):

        super(HornerMultivarPolynomial, self).__init__(coefficients, exponents, rectify_input, validate_input,
                                                       compute_representation)

        self.value_array_length = None
        self.representation = None

        # NOTE: do NOT automatically create all scalar factors with exponent 1
        # (they might be unused, since the polynomial must not actually depend on all variables)
        self.factor_container = FactorContainer()

        if find_optimal:
            self.factorisation_tree = OptimalFactorisationRoot(self.exponents, self.factor_container)
        else:
            self.factorisation_tree = HeuristicFactorisationRoot(self.exponents, self.factor_container)

        self.copy_recipe, self.scalar_recipe, self.monomial_recipe, self.tree_recipe, self.tree_ops = \
            self._compile_recipes()

        self.value_array_length = self.num_monomials + len(self.factor_container.scalar_factors) + len(
            self.factor_container.monomial_factors)

        self.root_value_idx = self.factorisation_tree.value_idxs[0]

        self.compute_string_representation(*args, **kwargs)  # uses num_ops!

        if not keep_tree:
            del self.factorisation_tree
            del self.factor_container

    def compute_string_representation(self, coeff_fmt_str: str = '{:.2}', factor_fmt_str: str = 'x_{dim}^{exp}', *args,
                                      **kwargs) -> str:
        representation = '[#ops={}] p(x)'.format(self.num_ops)
        if self.compute_representation:
            try:
                representation += ' = ' + self.factorisation_tree.get_string_representation(self.coefficients,
                                                                                            coeff_fmt_str,
                                                                                            factor_fmt_str)
                # exponentiation with 1 won't cause an operation in this representation
                # but are present in the string representation due to string formatting restrictions
                # -> they should not be displayed (misleading)
                representation = representation.replace('^1', '')  # <- workaround for the default string format
            except AttributeError:
                pass  # self.factorisation_tree does not exist

        self.representation = representation
        return self.representation

    def _compile_recipes(self):
        """ encode all instructions needed for evaluating the polynomial in 'recipes'

        recipes are represented as numpy ndarrays (cf. assembler instructions)

        -> acquire a data structure representing the factorisation tree
        -> avoid recursion and function call overhead during evaluation
        -> enables the use of jit compiled functions

        the factor container must now contain all unique factors used in the chosen factorisation
        during evaluation of a polynomial the values of all the factors are needed at least once
        -> compute the values of all factors once and store them
        -> store a pointer to the computed value for every factor ('value index' = address in the value array)
        this is required for compiling evaluation instructions depending on the factor values
        monomial factors exist only if their value is required during the evaluation of the parent polynomial
        scalar factors exist only if their value is required during the evaluation of existing monomial factors
        (scalar factors can be 'standalone' factors as well)
        -> values must not be overwritten (reusing addresses), because they might be needed again by another factor
        -> (without further optimisation) each factor requires its own space in the value array

        Returns:
            the compiled recipes (numpy ndarrays)
        """
        # the values of the factors are being stored after the coefficients
        # start the address assignment with the correct offset
        value_idx = self.num_monomials

        # compile the recipes for computing the value of all factors
        copy_recipe = []  # skip computing factors with exp 1, just copy x value
        scalar_recipe = []
        monomial_recipe = []

        # count the amount of multiplications encoded by the recipes
        # NOTE: count exponentiations as exponent-1 multiplications, e.g. x^3 <-> 2 operations
        self.num_ops = 0

        # -> IMPORTANT: value idx assignment must happen first for the scalar factors
        for scalar_factor in self.factor_container.scalar_factors:
            scalar_factor.value_idx = value_idx
            value_idx += 1
            copy_instr, scalar_instr = scalar_factor.get_recipe()
            copy_recipe += copy_instr
            scalar_recipe += scalar_instr
            if len(scalar_instr) > 0:
                exponent = scalar_instr[0][2]
                self.num_ops += exponent - 1

        for monomial_factor in self.factor_container.monomial_factors:
            monomial_factor.value_idx = value_idx
            value_idx += 1
            monomial_factor.factorisation_idxs = [f.value_idx for f in monomial_factor.scalar_factors]
            monomial_instr = monomial_factor.get_recipe()
            monomial_recipe += monomial_instr
            self.num_ops += 1  # every monomial instruction encodes one multiplication

        # compile the recipe for evaluating the Horner factorisation tree
        tree_recipe, tree_ops = self.factorisation_tree.get_recipe()
        # convert the recipes into the data types expected by the jit compiled functions
        # and store them
        tree_ops = np.array(tree_ops, dtype=BOOL_DTYPE)
        self.num_ops += len(tree_ops) - np.count_nonzero(tree_ops)  # every 0/False encodes a multiplication

        return (np.array(copy_recipe, dtype=UINT_DTYPE).reshape((-1, 2)),
                np.array(scalar_recipe, dtype=UINT_DTYPE).reshape((-1, 3)),
                np.array(monomial_recipe, dtype=UINT_DTYPE).reshape((-1, 3)),
                np.array(tree_recipe, dtype=UINT_DTYPE).reshape((-1, 2)),
                tree_ops)

    def eval(self, x: TYPE_1D_FLOAT, rectify_input: bool = False, validate_input: bool = False, ) -> float:
        """ computes the value of the polynomial at query point x

        makes use of fast ``Numba`` just in time compiled functions

        Args:
            x: ndarray of floats with shape = [self.dim] representing the query point
            rectify_input: bool, default=False
                whether to convert coefficients and exponents into compatible numpy arrays
                with this set to True, the query point x can be given in standard python arrays
            validate_input: check if the query point x fulfills the requirements

        Returns:
             the value of the polynomial at point x

        Raises:
            TypeError: if ``validate_input=True`` and x is not given as ndarray of dtype float
            ValueError: if ``validate_input=True`` and x does not have the shape ``[self.dim]``
        """

        if rectify_input:
            x = rectify_query_point(x)
        if validate_input:
            validate_query_point(x)
        if x.shape[0] != self.dim:
            raise ValueError(f'the query point x does not have the required dimensionality {self.dim}')

        value_array = np.empty(self.value_array_length, dtype=FLOAT_DTYPE)
        # the coefficients are being stored at the beginning of the value array
        # TODO remove flatten, always store coefficients as a 1D array (also for horner fact.)?!
        #   also in MultivarPolynomial.eval()
        value_array[:self.num_monomials] = self.coefficients.flatten()

        return eval_recipe(x, value_array, self.copy_recipe, self.scalar_recipe,
                           self.monomial_recipe, self.tree_recipe, self.tree_ops, self.root_value_idx)



# -*- coding:utf-8 -*-

# NOTE: if this raises SIGSEGV, update your Numba dependency

# TODO compare difference in computed values of other methods (=numerical error)
# TODO test all input parameter conversions, and data rectifications
# TODO test derivative correctness
# TODO test factorisation for 1D polynomials! should always find optimum = unique 1D Horner factorisation
# TODO test: addresses of factors must never be target addresses
#  (value would be overwritten, but might need to be reused)
#   might not be relevant in the future any more with other computation procedures reusing addresses


import itertools
import unittest

import numpy as np
import pytest

from multivar_horner.global_settings import FLOAT_DTYPE, UINT_DTYPE
from multivar_horner.multivar_horner import HornerMultivarPolynomial, MultivarPolynomial
# settings for numerical stability tests
from tests.test_helpers import naive_eval_reference, proto_test_case, vectorize
from tests.test_settings import (
    COEFF_CHANGE_DATA, INPUT_DATA_INVALID_TYPES_CONSTRUCTION, INPUT_DATA_INVALID_TYPES_QUERY,
    INPUT_DATA_INVALID_VALUES_CONSTRUCTION, INPUT_DATA_INVALID_VALUES_QUERY,
    MAX_INP_MAGNITUDE, NR_TEST_POLYNOMIALS, VALID_TEST_DATA,
)


class MainTest(unittest.TestCase):

    # TODO split up tests into cross class test cases and single class test cases,
    #  then use inheritance to change test class
    def test_construction_basic(self):
        """
        test the basic construction API functionalities
        :return:
        """
        print('\nTESTING COSNTRUCTION API...')
        coefficients = np.array([[5.0], [1.0], [2.0], [3.0]], dtype=FLOAT_DTYPE)
        exponents = np.array([[0, 0, 0], [3, 1, 0], [2, 0, 1], [1, 1, 1]], dtype=UINT_DTYPE)

        polynomial1 = MultivarPolynomial(coefficients, exponents, compute_representation=False)
        polynomial2 = MultivarPolynomial(coefficients, exponents, compute_representation=True)
        # both must have a string representation
        # [#ops=10] p(x)
        # [#ops=10] p(x) = 5.0 x_1^0 x_2^0 x_3^0 + 1.0 x_1^3 x_2^1 x_3^0 + 2.0 x_1^2 x_2^0 x_3^1 + 3.0 x_1^1 x_2^1 x_3^1
        assert len(str(polynomial1)) < len(str(polynomial2))
        assert str(polynomial1) == polynomial1.representation
        assert polynomial1.num_ops == 10

        return_str_repr = polynomial1.compute_string_representation(coeff_fmt_str='{:1.1e}',
                                                                    factor_fmt_str='(x{dim} ** {exp})')
        # the representation should get updated
        assert return_str_repr == polynomial1.representation

        polynomial1 = HornerMultivarPolynomial(coefficients, exponents, compute_representation=False)
        polynomial2 = HornerMultivarPolynomial(coefficients, exponents, compute_representation=True)
        r1 = str(polynomial1)  # [#ops=7] p(x)
        r2 = str(polynomial2)  # [#ops=7] p(x) = x_1 (x_1 (x_1 (1.0 x_2) + 2.0 x_3) + 3.0 x_2 x_3) + 5.0
        assert len(r1) < len(r2)
        assert r1 == polynomial1.representation
        assert r2 == polynomial2.representation
        assert polynomial1.num_ops == polynomial2.num_ops
        assert polynomial2.num_ops == 7

        return_str_repr = polynomial1.compute_string_representation(coeff_fmt_str='{:1.1e}',
                                                                    factor_fmt_str='(x{dim} ** {exp})')
        # the representation should get updated
        assert return_str_repr == polynomial1.representation

        # converting the input to the required numpy data structures
        coefficients = [5.0, 1.0, 2.0, 3.0]
        exponents = [[0, 0, 0], [3, 1, 0], [2, 0, 1], [1, 1, 1]]
        horner_polynomial = HornerMultivarPolynomial(coefficients, exponents, rectify_input=True, validate_input=True,
                                                     compute_representation=True, keep_tree=True)

        # search for an optimal factorisation
        horner_polynomial_optimal = HornerMultivarPolynomial(coefficients, exponents, find_optimal=True,
                                                             compute_representation=True, rectify_input=True,
                                                             validate_input=True)
        assert horner_polynomial_optimal.num_ops <= horner_polynomial.num_ops

        # partial derivative:
        deriv_2 = horner_polynomial.get_partial_derivative(2, compute_representation=True)

        # NOTE: partial derivatives themselves will be instances of the same parent class
        assert deriv_2.__class__ is horner_polynomial.__class__

        grad = horner_polynomial.get_gradient(compute_representation=True)
        # partial derivative for every dimension
        assert len(grad) == horner_polynomial.dim
        print('OK.\n')

    def test_invalid_input_detection(self):

        print('\n\nTEST INVALID INPUT DETECTION...')

        def construction_should_raise(data, expected_error):
            for inp, expected_output in data:
                coeff, exp, x = inp
                with pytest.raises(expected_error):
                    MultivarPolynomial(coeff, exp, rectify_input=False, validate_input=True, )
                with pytest.raises(expected_error):
                    HornerMultivarPolynomial(coeff, exp, rectify_input=False, validate_input=True, find_optimal=False)
                with pytest.raises(expected_error):
                    HornerMultivarPolynomial(coeff, exp, rectify_input=False, validate_input=True, find_optimal=True)

        construction_should_raise(INPUT_DATA_INVALID_TYPES_CONSTRUCTION, TypeError)
        construction_should_raise(INPUT_DATA_INVALID_VALUES_CONSTRUCTION, ValueError)

        # input rectification with negative exponents should raise a ValueError:
        coeff = [3.2]
        exp = [[0, 3, -4]]
        with pytest.raises(ValueError):
            HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=False, find_optimal=True)
        with pytest.raises(ValueError):
            MultivarPolynomial(coeff, exp, rectify_input=True, validate_input=False)
        with pytest.raises(ValueError):
            HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=False, find_optimal=False)

        def query_should_raise(data, expected_error):
            for inp, expected_output in data:
                coeff, exp, x = inp
                # NOTE: construction must not raise an error:
                p = MultivarPolynomial(coeff, exp, rectify_input=False, validate_input=True)
                with pytest.raises(expected_error):
                    p(x, rectify_input=False, validate_input=True)
                p = HornerMultivarPolynomial(coeff, exp, rectify_input=False, validate_input=True)
                with pytest.raises(expected_error):
                    p(x, rectify_input=False, validate_input=True)
                p = HornerMultivarPolynomial(coeff, exp, rectify_input=False, validate_input=True, find_optimal=True)
                with pytest.raises(expected_error):
                    p(x, rectify_input=False, validate_input=True)

        query_should_raise(INPUT_DATA_INVALID_TYPES_QUERY, TypeError)
        query_should_raise(INPUT_DATA_INVALID_VALUES_QUERY, ValueError)
        print('OK.')

    def test_eval_cases(self):
        def cmp_value_fct(inp):
            coeff, exp, x = inp
            x = np.array(x).T
            poly = MultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                      compute_representation=True)
            res1 = poly.eval(x, rectify_input=True, validate_input=False)
            print('MultivarPolynomial', poly)

            horner_poly = HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                                   compute_representation=True, find_optimal=False)
            res2 = poly.eval(x, rectify_input=True, validate_input=False)

            print('HornerMultivarPolynomial', horner_poly)
            # print('x=',x.tolist())
            horner_poly_opt = HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                                       compute_representation=True, find_optimal=True)
            res3 = poly.eval(x, rectify_input=True, validate_input=False)

            print('HornerMultivarPolynomial (optimal)', horner_poly_opt)
            # print('x=',x.tolist())

            if res1 != res2 or res2 != res3:
                print(f'x = {x}')
                print(f'results differ:\n{res1} (canonical)\n{res2} (horner)\n{res3} (horner optimal)\n')
                assert False

            assert horner_poly.num_ops >= horner_poly_opt.num_ops

            return res1

        print('\nTESTING EVALUATION CASES...')
        proto_test_case(VALID_TEST_DATA, cmp_value_fct)
        print('OK.\n')

    def test_eval_compare_vandermonde_nd(self):
        print('TEST COMPARISON TO VANDERMONDE POLYNOMIAL EVALUATION...')
        degree_range = range(1, 4)
        dim_range = range(1, 4)
        for dim, deg in itertools.product(dim_range, degree_range):
            exponents = np.array(list(itertools.product(range(deg), repeat=dim)), dtype=UINT_DTYPE)
            coeffs = np.random.rand(exponents.shape[0])
            X = np.random.rand(NR_TEST_POLYNOMIALS, dim)
            p_ref = naive_eval_reference(X, exponents, coeffs)
            for cls in [MultivarPolynomial, HornerMultivarPolynomial]:
                p_mv = vectorize(cls(coeffs[:, None], exponents))
                assert np.allclose(p_ref, p_mv(X))

        print('OK.\n')

    def test_eval_api(self):
        print('\nTESTING EVALUATION API...')
        keys = ['rectify_input', 'validate_input']
        vals = [True, False]
        max_dim = 3
        degree_range = range(1, 4)
        exponents = np.array(list(itertools.product(degree_range, repeat=max_dim)), dtype=UINT_DTYPE)
        coeffs = np.random.rand(exponents.shape[0])
        X = np.random.rand(NR_TEST_POLYNOMIALS, max_dim)
        p_ref = naive_eval_reference(X, exponents, coeffs)
        for kwargs in [dict(zip(keys, tf)) for tf in itertools.product(*[vals] * len(keys))]:
            for cls in [MultivarPolynomial, HornerMultivarPolynomial]:
                p_mv = vectorize(cls(coeffs[:, None], exponents, **kwargs))
                assert np.allclose(p_ref, p_mv(X))

        print('OK.\n')

    def test_eval_compare_numpy_1d(self):
        print('TEST COMPARISON TO NUMPY 1D POLYNOMIAL EVALUATION...')
        x = MAX_INP_MAGNITUDE * ((np.random.rand(NR_TEST_POLYNOMIALS) * 2) - 1)
        for ncoeff in range(1, 8):
            exponents = np.arange(ncoeff, dtype=UINT_DTYPE).reshape(ncoeff, 1)
            coeffs = np.random.rand(ncoeff)
            # p_np = lambda x: np.polyval(coeffs[::-1], x)
            p_np = np.polynomial.Polynomial(coeffs)
            for cls in [MultivarPolynomial, HornerMultivarPolynomial]:
                p_mv = vectorize(cls(coeffs[:, None], exponents))
                assert np.allclose(p_np(x), p_mv(x))

        print('OK.\n')

    def test_change_coefficients(self):

        print('\nTEST CHANGING COEFFICIENTS...')

        # Test if coefficients can actually be changed, representation should change accordingly

        def change_coeffs_fct(inp):
            print('\n')
            # changing the coefficients to the same coefficients should not alter the evaluation results
            # (reuse test data)

            coeffs1, exp, x, coeffs2 = inp
            # x = np.array(x).T
            print(x)
            poly = MultivarPolynomial(coeffs1, exp, rectify_input=True, validate_input=True,
                                      compute_representation=True)
            print(poly)
            poly = poly.change_coefficients(coeffs2, rectify_input=True, validate_input=True,
                                            compute_representation=True, )
            print(poly)
            res1 = poly.eval(x, rectify_input=True, validate_input=True)

            print('\n')
            horner_poly = HornerMultivarPolynomial(coeffs1, exp, rectify_input=True, validate_input=True,
                                                   compute_representation=True, keep_tree=True, find_optimal=False)
            print(horner_poly)
            horner_poly = horner_poly.change_coefficients(coeffs2, rectify_input=True, validate_input=True,
                                                          compute_representation=True, )
            res2 = horner_poly.eval(x, rectify_input=True, validate_input=True)
            print(horner_poly)
            # print('x=',x.tolist())

            print('\n')
            horner_poly_opt = HornerMultivarPolynomial(coeffs1, exp, rectify_input=True, validate_input=True,
                                                       compute_representation=True, keep_tree=True, find_optimal=True)
            print(horner_poly_opt)
            horner_poly_opt = horner_poly_opt.change_coefficients(coeffs2, rectify_input=True, validate_input=True,
                                                                  compute_representation=True, )
            res3 = horner_poly_opt.eval(x, rectify_input=True, validate_input=True)
            print(horner_poly_opt)
            # print('x=',x.tolist())

            if res1 != res2 or res2 != res3:
                print(f'x = {x}')
                print(f'results differ:\n{res1} (canonical)\n{res2} (horner)\n{res3} (horner optimal)\n')

            assert horner_poly.num_ops >= horner_poly_opt.num_ops

            return res1

        # this also tests if the factorisation tree, can still be accessed after changing the coefficients
        # representation would otherwise be empty
        # keep_tree has to be True
        proto_test_case(COEFF_CHANGE_DATA, change_coeffs_fct)

        def cmp_value_changed_coeffs_fct(inp):
            # changing the coefficients to the same coefficients should not alter the evaluation results
            # (reuse test data)
            coeff, exp, x = inp
            # x = np.array(x).T
            poly = MultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                      compute_representation=True)
            poly = poly.change_coefficients(coeff, rectify_input=True, validate_input=True,
                                            compute_representation=True, )
            print(poly)
            res1 = poly.eval(x, rectify_input=True, validate_input=True)

            horner_poly = HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                                   compute_representation=True, find_optimal=False, keep_tree=True)
            horner_poly = horner_poly.change_coefficients(coeff, rectify_input=True, validate_input=True,
                                                          compute_representation=True, )
            print(horner_poly)
            res2 = horner_poly.eval(x, rectify_input=True, validate_input=True)
            # print('x=',x.tolist())

            horner_poly_opt = HornerMultivarPolynomial(coeff, exp, rectify_input=True, validate_input=True,
                                                       compute_representation=True, keep_tree=True, find_optimal=True)
            horner_poly_opt = horner_poly_opt.change_coefficients(coeff, rectify_input=True, validate_input=True,
                                                                  compute_representation=True, )
            print(horner_poly_opt)
            res3 = horner_poly_opt.eval(x, rectify_input=True, validate_input=True)
            # print('x=',x.tolist())

            if res1 != res2 or res2 != res3:
                print(f'x = {x}')
                print(f'results differ:\n{res1} (canonical)\n{res2} (horner)\n{res3} (horner optimal)\n')
                assert False

            assert horner_poly.num_ops >= horner_poly_opt.num_ops

            return res1

        proto_test_case(VALID_TEST_DATA, cmp_value_changed_coeffs_fct)
        print('OK.\n')


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(MainTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
    unittest.main()
